"""
fsspec wrappers that should make working with S3 / the local file system
seemless.

TODO:
    Someone must have already implemented this somewhere. Find that to
    either use directly or as a reference.

Look into:
    https://github.com/fsspec/universal_pathlib
    https://pypi.org/project/pathlibfs/
"""
from os.path import join
import pathlib
import ubelt as ub
import os
import fsspec
# import types

NOOP_CALLBACK = fsspec.callbacks.NoOpCallback()


class FSPath(str):
    """
    Provide a pathlib.Path-like way of interacting with fsspec.

    This has a few notable differences with pathlib.Path. We inherit from
    :class:`str` because :class:`pathlib.Path` semantics can break protocols
    sections of URIs. This means we have to use :mod:`os.path` functions
    to implement things like :meth:`FSPath.relative_to` and
    :meth:`FSPath.joinpath` (which behave differently than pathlib)

    Note:
        Not all of the fsspec / pathlib operations are currently implemented,
        add as needed.

    Example:
        >>> cwd = FSPath.coerce('.')
        >>> print(cwd)
        >>> print(cwd.fs)
    """
    # Final subclasses must define this as a string to be passed to
    # fsspec.filesystem(__protocol__)
    # __protocol__ : str | types.NotImplementedType = NotImplemented
    __protocol__ = NotImplemented

    @classmethod
    def _new_fs(cls, **kwargs):
        """
        Create a new filesystem instance based on __protocol__
        """
        fs_cls = cls.get_filesystem_class()
        fs = fs_cls(**kwargs)
        return fs

    @classmethod
    def get_filesystem_class(cls):
        fs_cls = fsspec.get_filesystem_class(cls.__protocol__)
        return fs_cls

    @classmethod
    def _current_fs(cls):
        """
        The "default" FileSystem object.  Get the most recent filesystem with
        this protocol, or create a new one with defaults.

        Returns:
            AbstractFileSystem
        """
        fs_cls = cls.get_filesystem_class()
        fs = fs_cls.current()
        return fs

    def __new__(cls, path, *, fs=None):
        # Note: the value of the string is set in the __new__ method because
        # strings are immutable. So we dont need to call super or anything.
        # The first argument will be the value of the string.
        if fs is None:
            # Lazy creation of a new fs
            fs = cls._current_fs()
        self = str.__new__(cls, path)
        self._fs: fsspec.AbstractFileSystem = fs
        return self

    @property
    def fs(self) -> fsspec.AbstractFileSystem:
        return self._fs

    @fs.setter
    def fs(self, value: fsspec.AbstractFileSystem):
        self._fs = value

    @classmethod
    def coerce(cls, path):
        """
        Determine which backend to use automatically

        Example:
            >>> path2 = FSPath.coerce('/local/path')
            >>> print(f'path2={path2}')
            >>> assert path2.is_local()
            >>> # xdoctest: +REQUIRES(module:s3fs)
            >>> path1 = FSPath.coerce('s3://demo_bucket')
            >>> print(f'path1={path1}')
            >>> assert path1.is_remote()
        """
        path = os.fspath(path)
        if path.startswith(('s3:', '/vsis3/')):
            self = S3Path.coerce(path)
        elif path.startswith(('ssh:')):
            raise NotImplementedError('getting ssh to work with uris is tricky')
        else:
            self = LocalPath(path)
        return self

    @property
    def path(self):
        """
        By default the string representation is assumed to be the entire path,
        however, for subclasses like SSHPath it is necessary to overwrite this
        so the core object represents the entire URI, but this just returns the
        path part, which is what the fsspec.FileSystem object expects.
        """
        return self

    def relative_to(self, other):
        return self.__class__(os.path.relpath(self, other))

    def is_remote(self):
        return isinstance(self, RemotePath)

    def is_local(self):
        return isinstance(self, LocalPath)

    def open(self, mode='rb', block_size=None, cache_options=None, compression=None):
        """
        Example:
            >>> from geowatch.utils.util_fsspec import *  # NOQA
            >>> from geowatch.utils import util_fsspec
            >>> dpath = util_fsspec.LocalPath.appdir('geowatch/fsspec/tests/open').ensuredir()
            >>> fpath = dpath / 'file.txt'
            >>> file = fpath.open(mode='w')
            >>> file.write('hello world')
            >>> file.close()
            >>> assert fpath.read_text() == fpath.open('r').read()
        """
        return self.fs.open(self.path, mode=mode, block_size=block_size,
                            cache_options=cache_options,
                            compression=compression)

    def ls(self, detail=False, **kwargs):
        """
        Example:
            >>> from geowatch.utils.util_fsspec import *  # NOQA
            >>> import ubelt as ub
            >>> dpath = ub.Path.appdir('geowatch', 'tests', 'fsspec', 'ls').ensuredir()
            >>> (dpath / 'file1').touch()
            >>> (dpath / 'file2').touch()
            >>> (dpath / 'subfolder').ensuredir()
            >>> self = FSPath.coerce(dpath)
            >>> results = self.ls()
            >>> assert sorted(results) == sorted(map(str, dpath.ls()))
        """
        cls = self.__class__
        results = self.fs.ls(self.path, detail=detail, **kwargs)
        if detail:
            return results
        else:
            # Hack:
            if self.__protocol__ == 'file':
                return [cls(p, fs=self.fs) for p in results]
            else:
                return [cls(self.__protocol__ + '://' + p, fs=self.fs) for p in results]

    def touch(self, truncate=False, **kwargs):
        """
        Example:
            >>> import ubelt as ub
            >>> dpath = ub.Path.appdir('geowatch', 'tests', 'fsspec', 'touch').ensuredir()
            >>> dpath_ = FSPath.coerce(dpath)
            >>> self = (dpath_ / 'file')
            >>> self.touch()
            >>> assert self.exists()
            >>> assert (dpath / 'file').exists()
        """
        self.fs.touch(self.path, truncate=truncate, **kwargs)

    def move(self, path2, recursive='auto', maxdepth=None, idempotent=True, verbose=1, **kwargs):
        """
        Note: this may work differently than ubelt.Path.move, ideally we should
        rectify this. The difference case is what happens when you move:

            ./path/to/dir -> ./path/to/other/dir

        Does `./path/to/dir` merge into `./path/to/other/dir`, or do you get
        all of the src contents in `./path/to/other/dir/dir`?

        Ignore:
            import ubelt as ub
            root = ub.Path.appdir('geowatch', 'tests', 'fsspec', 'move').delete().ensuredir()
            dpath1 = (root / 'path/to/dir').ensuredir()
            dpath2 = (root / 'path/to/other/dir').ensuredir()

            # Add content to both dirs
            (dpath1 / 'file1').write_text('a1')
            (dpath1 / 'file2').write_text('a2')
            (dpath2 / 'file1').write_text('b1')
            (dpath2 / 'file3').write_text('b3')

            dpath1.ls()
            dpath2.ls()

            # ubelt will complain moves are only allowed to locs that dont
            # exist cool, that makes sense.
            dpath1.move(dpath2)

            dpath1_alt = FSPath.coerce(dpath1)
            dpath2_alt = FSPath.coerce(dpath2)

            if 0:
                # This moves it into the directory, which is not the behavior I
                # want.
                dpath1_alt.move(dpath2_alt)
                assert not dpath1.exists()
                dpath2.ls()
            else:
                # This has the desired behavior, but requries an akward call
                # sig
                dpath1_alt.move(dpath2_alt.parent)
                assert not dpath1.exists()
                dpath2.ls()
        """
        if recursive == 'auto':
            recursive = self.is_dir()
        if verbose:
            print(f'Move {self} -> {path2}')

        src_path = self.path
        dst_path = path2

        if idempotent:
            # Ensure that if we're copying directories that the paths have
            # a trailing slash, otherwise when copying a directory into a
            # non-empty directory the source directory itself (rather than
            # just the contents) will be copied into the destination
            # directory
            if self.is_dir():
                if not self.path.endswith('/'):
                    src_path = f"{self.path}/"

                if not path2.endswith('/'):
                    dst_path = f"{path2}/"

        self.fs.move(src_path, dst_path, recursive=recursive, maxdepth=maxdepth,
                     **kwargs)

    def delete(self, recursive='auto', maxdepth=True, verbose=1):
        """
        Deletes this file or this directory (and all of its contents)

        Unlike fs.delete, this will not error if the file doesnt exist. See
        :func:`FSPath.rm` if you want standard error-ing behavior.
        """
        if verbose:
            print(f'Delete {self}')
        if recursive == 'auto':
            recursive = self.is_dir()
        try:
            return self.fs.delete(self.path, recursive=recursive, maxdepth=maxdepth)
        except FileNotFoundError:
            ...

    def rm(self, recursive='auto', maxdepth=True):
        """
        Deletes this file or this directory (and all of its contents)
        """
        if recursive == 'auto':
            recursive = self.is_dir()
        return self.fs.rm(self.path, recursive=recursive, maxdepth=maxdepth)

    def mkdir(self, create_parents=True, **kwargs):
        """
        Note:
            does nothing on some filesystems (e.g. S3)
        """
        return self.fs.mkdir(self.path, create_parents=create_parents, **kwargs)

    def stat(self):
        return self.fs.stat(self.path)

    def is_dir(self):
        return self.fs.isdir(self.path)

    def is_file(self):
        return self.fs.isfile(self.path)

    def is_link(self):
        try:
            return self.fs.islink(self.path)
        except AttributeError:
            return False

    def exists(self):
        return self.fs.exists(self.path)

    def write_text(self, value, **kwargs):
        return self.fs.write_text(self.path, value, **kwargs)

    def read_text(self, **kwargs):
        return self.fs.read_text(self.path, **kwargs)

    def walk(self, include_protocol='auto', **kwargs):
        """
        Yields:
            Tuple[Self, List[str], List[str]] - root, dir names, file names
        """
        if include_protocol == 'auto':
            include_protocol = self.is_remote()
        if include_protocol:
            for root, dnames, fnames in self.fs.walk(self.path, **kwargs):
                root = self.__class__(self.fs.unstrip_protocol(root), fs=self.fs)
                yield root, dnames, fnames
        else:
            for root, dnames, fnames in self.fs.walk(self, **kwargs):
                root = self.__class__(root, fs=self.fs)
                yield root, dnames, fnames

    @property
    def parent(self):
        """
        Example:
            >>> self = FSPath.coerce('foo/bar/baz.jaz.raz')
            >>> assert str(ub.Path(self).parent) == self.parent
            >>> assert self.parent == 'foo/bar'
        """
        return self.__class__(os.path.dirname(self), fs=self.fs)

    @property
    def name(self):
        """
        Example:
            >>> self = FSPath.coerce('foo/bar/baz.jaz')
            >>> assert ub.Path(self).name == self.name
            >>> assert self.name == 'baz.jaz'
        """
        return os.path.basename(self.path)

    @property
    def stem(self):
        """
        Example:
            >>> self = FSPath.coerce('foo/bar/baz.jaz')
            >>> assert ub.Path(self).stem == self.stem
            >>> assert self.stem == 'baz'
        """
        return os.path.splitext(self.path.name)[0]

    @property
    def suffix(self):
        """
        Example:
            >>> self = FSPath.coerce('foo/bar/baz.jaz.raz')
            >>> assert ub.Path(self).suffix == self.suffix
            >>> assert self.suffix == '.raz'
        """
        return os.path.splitext(self.path.name)[1]

    @property
    def suffixes(self):
        """
        Example:
            >>> self = FSPath.coerce('foo/bar/baz.jaz.raz')
            >>> assert ub.Path(self).suffixes == self.suffixes
            >>> assert self.suffixes == ['.jaz', '.raz']
        """
        return ['.' + x for x in self.name.split('.')[1:]]

    @property
    def parts(self):
        """
        Example:
            >>> self = FSPath.coerce('foo/bar/baz.jaz.raz')
            >>> assert ub.Path(self).parts == self.parts
            >>> assert self.parts == ('foo', 'bar', 'baz.jaz.raz')
        """
        return tuple(self.path.split(self.fs.sep))

    def copy(self, dst, recursive='auto', maxdepth=None, on_error=None,
             callback=None, verbose=1, idempotent=True, overwrite=False,
             **kwargs):
        """
        Copies this file or directory to dst

        Abtracts fsspec copy / put / get.

        If dst ends with a "/", it will be assumed to be a directory, and
        target files will go within.

        Unlike fsspec, this attempts to be idempotent. See [FSSpecCopy]_.

        Args:
            dst (FSPath): location to copy to

            recursive (bool | str):
                If 'auto' (the default), attempt to determine if this is a
                directory or a file. Set to True if it is a directory and False
                otherwise. If you know what this is beforehand, you can set it
                explicitly to be more efficient.

            maxdepth (int | None):
                only makes sense when recursive is True

            callback (None | callable):
                for put / get cases

            on_error (str):
                either "raise", "ignore". Only applicable in the "copy" case.

            idempotent (bool):
                if False, use standard fsspec behavior, otherwise attempt to
                be idempotent.

            overwrite (bool):
                if True, overwrite existing data instead of erroring. Defaults
                to False.

        Note:
            There are different functions depending on if we are going from
            remote->remote (copy), local->remote (put), or remote->local (get)

        References:
            .. [FSSpecCopy] https://filesystem-spec.readthedocs.io/en/latest/copying.html

        Example:
            >>> from geowatch.utils import util_fsspec
            >>> dpath = util_fsspec.LocalPath.appdir('geowatch/fsspec/tests/copy').ensuredir()
            >>> src_dpath = (dpath / 'src').ensuredir()
            >>> for i in range(100):
            ...     (src_dpath / 'file_{i:03d}.txt').write_text('hello world' * 100)
            >>> dst_dpath = (dpath / 'dst')
            >>> dst_dpath.delete()
            >>> src_dpath.copy(dst_dpath, verbose=3)
            >>> dst_dpath.delete()
            >>> if 0:
            >>>     from fsspec.callbacks import TqdmCallback
            >>>     callback = TqdmCallback(tqdm_kwargs={"desc": "Your tqdm description"})
            >>>     src_dpath.copy(dst_dpath, callback=callback)
        """
        if recursive == 'auto':
            # On S3, asking if something is a dir can give permission
            # issues, whereas asking if something is a file seems ok.
            # try:
            # recursive = self.is_dir()
            # except PermissionError:
            recursive = not self.is_file()

        if callback is None:
            callback = NOOP_CALLBACK

        commonkw = {
            'recursive': recursive,
            'maxdepth': maxdepth,
            **kwargs,
        }

        dst: FSPath

        if overwrite:
            raise NotImplementedError

        if verbose:
            print(f'Copy {self} -> {dst}')

        # HANDLE SPECIAL CASE WHERE FSSPEC IS NOT IDEMPOTENT
        if idempotent:
            if recursive and dst.exists():
                dst = dst.parent + '/'

        if isinstance(self, LocalPath):
            if not self.exists():
                raise IOError(f'{self} does not exist')
            if isinstance(dst, RemotePath):

                # TODO: test if we are an empty directory and fail because
                # generally we cant copy an empty directory to a remote.
                try:
                    if recursive:
                        if verbose >= 3:
                            print(' * local -> remote (put recursive)')
                        return dst.fs.put(self.path, dst, **commonkw, callback=callback)
                    else:
                        if verbose >= 3:
                            print(' * local -> remote (put_file)')
                        return dst.fs.put_file(self.path, dst, callback=callback)
                except FileExistsError:
                    # TODO: overwrite
                    raise

            elif isinstance(dst, LocalPath):
                if verbose >= 3:
                    print(' * local -> local')
                return self.fs.copy(self.path, dst, **commonkw, callback=callback)
            else:
                raise TypeError(type(dst))
        elif isinstance(self, RemotePath):
            if isinstance(dst, RemotePath):
                if verbose >= 3:
                    print(' * remote -> remote')
                return self.fs.copy(self.path, dst, **commonkw, on_error=on_error)
            elif isinstance(dst, (LocalPath, pathlib.Path)):

                if recursive:
                    if verbose >= 3:
                        print(' * remote -> local (get recursive)')
                    return self.fs.get(self.path, dst, **commonkw, callback=callback)
                else:
                    # Using put on an s3 bucket seems like it fails when
                    # directories have tight permissions, but put file
                    # seems ok. Need to ensure that we are actually just
                    # using a file in this instance.
                    if verbose >= 3:
                        print(' * remote -> local (get_file)')
                    return self.fs.get_file(self.path, dst, callback=callback)
            else:
                raise TypeError(type(dst))
        else:
            raise TypeError(type(self))

    def joinpath(self, *others):
        return self.__class__(join(self, *others), fs=self.fs)

    def __truediv__(self, other):
        return self.__class__(join(self, other), fs=self.fs)

    def __add__(self, other):
        """
        Returns a new string starting with this fspath representation.
        """
        return self.__class__(str(self) + other, fs=self.fs)

    def __radd__(self, other):
        """
        Returns a new string ending with this fspath representation.
        """
        return self.__class__(other + str(self), fs=self.fs)

    def tree(self, max_files=100, dirblocklist=None, show_nfiles='auto',
             return_text=False, return_tree=True, pathstyle='name',
             max_depth=None, with_type=False, abs_root_label=True,
             colors=not ub.NO_COLOR):
        """
        Filesystem tree representation

        Like the unix util tree, but allow writing numbers of files per directory
        when given -d option

        Ported from xdev.misc.tree_repr

        TODO:
            instead of building the networkx structure and then waiting to
            display everything, build and display simultaniously. Will require
            using a modified version of write_network_text

        Args:
            max_files (int | None) : maximum files to print before supressing a directory
            pathstyle (str): can be rel, name, or abs
            return_tree (bool): if True return the tree
            return_text (bool): if True return the text
            maxdepth (int | None): maximum depth to descend
            abs_root_label (bool): if True force the root to always be absolute
            colors (bool): if True use rich
        """
        import io
        import os
        import networkx as nx
        from cmd_queue.util.util_networkx import write_network_text
        from kwutil.util_pattern import MultiPattern
        # tree = nx.OrderedDiGraph()
        tree = nx.DiGraph()

        if dirblocklist is not None:
            dirblocklist = MultiPattern.coerce(dirblocklist, hint='glob')

        def _make_label(p, force_abs=False):
            if force_abs:
                pathrep = p
            elif pathstyle == 'rel':
                pathrep = p.relative_to(self)
            elif pathstyle == 'name':
                pathrep = p.name
            elif pathstyle == 'abs':
                pathrep = p
            else:
                KeyError(pathstyle)

            types = []
            islink = p.is_link()
            isdir = p.is_dir()
            isfile = p.is_file()
            isbroken = False
            scolor = ''
            tcolor = ''
            L_scolor = ''
            L_tcolor = ''
            if islink:
                if colors:
                    L_scolor = '[cyan]'
                    L_tcolor = '[/cyan]'
                types.append('L')
                if not isfile and not isdir:
                    isbroken = True
                    if isbroken:
                        if colors:
                            L_scolor = '[red]'
                            L_tcolor = '[/red]'
                    types.append('B')

            if isfile:
                if colors:
                    scolor = '[reset]'
                    tcolor = '[/reset]'
                    if os.access(p, os.X_OK):
                        scolor = '[green]'
                        tcolor = '[/green]'
                types.append('F')
            if isdir:
                if colors:
                    scolor = f'[blue][link={p.absolute()}]'
                    tcolor = '[/link][/blue]'
                types.append('D')

            if islink:
                target = os.readlink(p)
                pathrep = L_scolor + pathrep + L_tcolor + ' -> ' + scolor + target + tcolor
            else:
                pathrep = scolor + pathrep + tcolor

            if with_type:
                typelabel = ''.join(types)
                return f'({typelabel}) ' + pathrep
            else:
                return pathrep

        # TODO: rectify with "find"
        start_depth = self.count(self.fs.sep)
        for root, dnames, fnames in self.walk():
            curr_depth = root.count(self.fs.sep)

            if max_depth is not None:
                if (curr_depth - start_depth):
                    del dnames[:]

            if dirblocklist is not None:
                dnames[:] = [
                    dname for dname in dnames if not dirblocklist.match(dname)]

            dnames[:] = sorted(dnames)
            tree.add_node(root)

            too_many_files = max_files is not None and len(fnames) >= max_files

            if show_nfiles == 'auto':
                show_nfiles_ = too_many_files
            else:
                show_nfiles_ = show_nfiles

            num_files = len(fnames)
            if show_nfiles_:
                prefix = '[ {} ] '.format(num_files)
            else:
                prefix = ''

            force_abs = abs_root_label and curr_depth == start_depth
            label = '{}{}'.format(prefix, _make_label(root, force_abs))

            tree.nodes[root]['label'] = label

            if not too_many_files:
                for fname in fnames:
                    fpath = root / fname
                    tree.add_node(fpath)
                    tree.nodes[fpath]['label'] = _make_label(fpath)
                    tree.add_edge(root, fpath)

            for dname in dnames:
                dpath = root / dname
                tree.add_node(dpath)
                tree.nodes[dpath]['label'] = _make_label(dpath)
                tree.add_edge(root, dpath)

        file = io.StringIO()
        write_network_text(tree, file)
        file.seek(0)
        text = file.read()

        info = {}

        if return_text:
            info['text'] = text
        else:
            if colors:
                from rich import print as rprint
                rprint(text)
            else:
                print(text)

        if return_tree:
            info['tree'] = tree
        return info


class LocalPath(FSPath):
    """
    The implementation for the local filesystem

    CommandLine:
        xdoctest -m geowatch.utils.util_fsspec LocalPath
        xdoctest geowatch/utils/util_fsspec.py

    Example:
        >>> from geowatch.utils.util_fsspec import *  # NOQA
        >>> dpath = ub.Path.appdir('geowatch/tests/util_fsspec/demo')
        >>> dpath.delete().ensuredir()
        >>> (dpath / 'file1.txt').write_text('data')
        >>> (dpath / 'dpath').ensuredir()
        >>> (dpath / 'dpath/file2.txt').write_text('data')
        >>> self = LocalPath(dpath).absolute()
        >>> print(f'self={self}')
        >>> print(self.ls())
        >>> info = self.tree()
        >>> fsspec_dpath = (dpath / 'dpath')
        >>> fsspec_fpath = (dpath / 'file1.txt')
        >>> pathlib_dpath = ub.Path(dpath / 'pathlib_dpath')
        >>> pathlib_fpath = ub.Path(dpath / 'pathlib_fpath')
        >>> assert not pathlib_dpath.exists()
        >>> assert not pathlib_fpath.exists()
        >>> fsspec_dpath.copy(pathlib_dpath)
        >>> fsspec_fpath.copy(pathlib_fpath)
        >>> assert pathlib_dpath.exists()
        >>> assert pathlib_fpath.exists()
    """
    __protocol__ = 'file'

    def ensuredir(self, mode=0o0777):
        pathlib.Path(self).mkdir(mode=mode, parents=True, exist_ok=True)
        return self

    def absolute(self):
        return self.__class__(os.path.abspath(self), fs=self.fs)

    @classmethod
    def appdir(cls, *args, **kw):
        return cls(str(ub.Path.appdir(*args, **kw)))


class MemoryPath(FSPath):
    """
    Ignore:
        self = MemoryPath('/')
        self.ls()
        (self / 'file').write_text('data')
        self.ls()

        ref_mempath = MemoryPath('/')
        ref_mempath.ls()

        # The MemoryFileSystem is global for the entire program
        new_mempath = MemoryPath('/', fs=MemoryPath._new_fs())
        new_mempath.ls()

        # Show the entire contents of the memory filesystem
        new_mempath.fs.store
    """
    __protocol__ = 'memory'


class RemotePath(FSPath):
    """
    Abstract implementation for all remote filesystems
    """


class S3Path(RemotePath):
    """
    The specific S3 remote filesystem.

    Control credentials with the environment variables: AWS_ACCESS_KEY_ID,
    AWS_SECRET_ACCESS_KEY, and AWS_SESSION_TOKEN.

    A single S3 filesystem is used by default, but you can work with multiple
    of them if you pass in the fs object. E.g.

        fs = S3Path._new_fs(profile='iarpa')
        self = S3Path('s3://kitware-smart-watch-data/', fs=fs)
        self.ls()

        # Can also do
        S3Path.register_bucket('s3://kitware-smart-watch-data', profile='iarpa')
        self = S3Path.coerce('s3://kitware-smart-watch-data/')
        self.ls()

        # Demo of multiple registered buckets
        S3Path.register_bucket('s3://usgs-landsat-ard', profile='iarpa', requester_pays=True)
        self = S3Path.coerce('s3://usgs-landsat-ard/collection02')
        self.ls()

        self = S3Path.coerce('/vsis3/usgs-landsat-ard/collection02')
        self.ls()

    SeeAlso:
        geowatch.heuristics.register_known_fsspec_s3_buckets

    To work with different S3 filesystems,

    See [S3FS_Docs]_.

    Requirements:
        s3fs>=2023.6.0

    References:
        .. [S3FS_Docs] https://s3fs.readthedocs.io/en/latest/?badge=latest

    Example:
        >>> # xdoctest: +REQUIRES(module:s3fs)
        >>> fs = S3Path._new_fs()
    """
    __protocol__ = 's3'

    _bucket_registry = {}

    def _as_gdal_vsi(self):
        # replace the s3:// part with /vsis3
        return '/vsis3/' + self[len(self.__protocol__) + 3:]

    def ensuredir(self, mode=0o0777):
        # Does nothing on S3 because you cannot create an empty directory
        # they are always auto-created for you.
        ...
        return self

    @classmethod
    def register_bucket(cls, bucket, **kwargs):
        # Parse out the bucket part
        assert '/' not in bucket[5], 'not a bucket'
        fs = cls._new_fs(**kwargs)
        # want to store and associate buckets with fs objects
        cls._bucket_registry[bucket] = fs

    @classmethod
    def coerce(cls, path):
        if path.startswith('/vsis3/'):
            # convert gdal virtual filesystems to s3 paths
            path = 's3://' + path[7:]
        # Parse out the bucket part
        bucket = path[:path.find('/', 5)]
        fs = cls._bucket_registry.get(bucket, None)
        self = cls(path, fs=fs)
        return self


class SSHPath(RemotePath):
    """
    Ignore:
        fs = SSHPath._new_fs(host='localhost')
        self = SSHPath('.', fs=fs)
        self.ls()
        self = SSHPath('misc/notes', fs=fs)
        self.tree()
    """
    __protocol__ = 'ssh'

    @property
    def host(self):
        return self.fs.host


class _BROKEN_SSHURI(RemotePath):
    """
    The idea here is to do something like the bucket registry in S3Path, but
    weird corner cases pop up here making the path forward not easy to
    identify.

    Requires a RFC3986 style URI.

    The idea here is that we keep the remote information in the string so we
    can automatically lookup the appropriate fs object and only have to
    authenticate it once.

    Note this is a mess, and we may just drop this idea...

    Ignore:
        from geowatch.utils.util_fsspec import *  # NOQA
        self = SSHPath('ssh://localhost/.')
        print(f'self={self}')
        self.ls()
        (self / 'misc/notes').tree()

        self = SSHPath('misc/notes', fs=fs)
        self.tree()

        fs = SSHPath._new_fs(host='localhost')
        path = 'ssh://localhost/.'
        self = SSHPath(path, fs=fs)
    """
    __protocol__ = 'ssh'

    _fs_registry = {}

    @property
    def host(self):
        return self.fs.host

    @classmethod
    def _lookup_fs_remote(cls, path, fs=None, **kwargs):
        import uritools
        uri_parsed = uritools.urisplit(path)

        remote = uritools.uricompose(
            scheme=uri_parsed.scheme, authority=uri_parsed.authority,
            userinfo=uri_parsed.userinfo, host=uri_parsed.host,
            port=uri_parsed.port)

        if remote in cls._fs_registry:
            fs = cls._fs_registry[remote]

        if fs is None:
            fs = cls.register_remote(remote, **kwargs)

        return fs

    @classmethod
    def register_remote(cls, remote, **kwargs):
        import uritools
        uri_parsed = uritools.urisplit(remote)
        assert not uri_parsed.path
        remote = uritools.uricompose(
            scheme=uri_parsed.scheme, authority=uri_parsed.authority,
            userinfo=uri_parsed.userinfo, host=uri_parsed.host,
            port=uri_parsed.port)
        ssh_kwargs = {}
        ssh_kwargs['host'] = uri_parsed.host
        if uri_parsed.port is not None:
            ssh_kwargs['port'] = uri_parsed.port
        if uri_parsed.userinfo is not None:
            ssh_kwargs['username'] = uri_parsed.userinfo
        ssh_kwargs.update(kwargs)
        print('new remote')
        print(f'remote={remote}')
        print(f'ssh_kwargs={ssh_kwargs}')
        fs = cls._new_fs(**ssh_kwargs)
        print(f'fs={fs}')
        cls._fs_registry[remote] = fs
        return fs

    def __new__(cls, path, *, fs=None):
        parsed, remote = _uri_parse(path)
        if fs is None:
            fs = cls._lookup_fs_remote(path)
        self = super().__new__(cls, path, fs=fs)
        self._path = parsed.path
        return self

    @property
    def path(self):
        return self._path


def _uri_parse(uri):
    import uritools
    uri_parsed = uritools.urisplit(uri)
    remote = uritools.uricompose(
        scheme=uri_parsed.scheme, authority=uri_parsed.authority,
        userinfo=uri_parsed.userinfo, host=uri_parsed.host,
        port=uri_parsed.port)
    return uri_parsed, remote


def _devtest_ssh_registry1():
    """
    TODO:
        - [ ] Generate a test that setups the docker ssh server

    This assumes you've run the instructions in
        ~/code/ci-docker/ubuntu_sshd.dockerfile

    and have a local ssh server running
    """
    from geowatch.utils.util_fsspec import SSHPath

    ssh_kwargs = {
        'port': 2222,
        'username': 'ubuntu',
        # 'password': 'ubuntu',
        'key_filename': str(ub.Path('~/code/ci-docker/tmp_keys/id_ed25519').expand()),
    }
    fs1 = SSHPath._new_fs(host='localhost', **ssh_kwargs)

    fs2 = SSHPath._new_fs(host='localhost')
    fs1.open('testfile2', mode='w').write('foo')

    # Should be files on the docker ssh server
    fs1.ls('.')

    # Should be files on the local ssh server (todo: test with 2 docker instances)
    fs2.ls('.')

    # import fsspec
    # uri1 = 'ssh://localhost:22/testfile1'
    # uri2 = 'ssh://ubuntu@localhost:2222/testfile2'
    # file1 = fsspec.open(uri1, mode='w')
    # file2 = fsspec.open(uri2)

    # p1 = SSHPath(uri1, fs=fs1)
    # p2 = SSHPath(uri1, fs=fs1)
    # fsspec.core.get_fs_token_paths(uri1)
    # fsspec.core.get_fs_token_paths(uri2)

    # # uri1 = 'ssh://localhost:22/.ssh'
    # uri2 = 'ssh://ubuntu@localhost:2222/.ssh'

    # import uritools
    # # uri_parsed1 = uritools.urisplit(uri1)
    # # uri_parsed2 = uritools.urisplit(uri2)

    # remote = uri2
    # options = {
    #     'key_filename': str(ub.Path('~/code/ci-docker/tmp_keys/id_ed25519').expand()),
    # }
    # registry = {}
    # if remote in registry:
    #     raise ValueError('already exists')
    # uri_parsed = uritools.urisplit(remote)
    # options['port'] = uri_parsed.port
    # options['host'] = uri_parsed.host
    # options['username'] = uri_parsed.userinfo
    # fs = SSHPath._new_fs(**options)
    # # registry
    # prefix = uritools.uricompose(
    #     scheme=uri_parsed.scheme, authority=uri_parsed.authority,
    #     userinfo=uri_parsed.userinfo, host=uri_parsed.host,
    #     port=uri_parsed.port)
    # registry[prefix] = fs

    # uri = 's3://usgs-landsat-ard/collection02/oli-tirs/2016/CU/003/008/LC08_CU_003008_20161026_20210502_02/LC08_CU_003008_20161026_20210502_02_SR_B2.TIF'
    # parsed, prefix = _uri_parse(uri)
    # prefix = 's3://usgs-landsat-ard'
    # parsed, prefix = _uri_parse(uri)

    from uritools import uricompose
    from uritools import urisplit
    uricompose(scheme='foo', host='example.com', port=8042, path='/over/there',
               query={'name': 'ferret'}, fragment='nose')

    x = uricompose(scheme='foo', host='example.com', port=8042, path='/.')
    urisplit(x).authority
    uricompose(scheme='foo', host='example.com', path='/.')


def _devtest_ssh_registry2():
    # from geowatch.utils.util_fsspec import SSHPath, FSPath
    prefix1 = 'ssh://localhost:22'
    prefix2 = 'ssh://ubuntu@localhost:2222'
    options1 = {
        'port': 22,
        'host': 'localhost',
    }
    options2 = {
        'port': 2222,
        'host': 'localhost',
        'username': 'ubuntu',
        'key_filename': str(ub.Path('~/code/ci-docker/tmp_keys/id_ed25519').expand()),
    }
    f1 = SSHPath.register_remote(prefix1, **options1)
    f2 = SSHPath.register_remote(prefix2, **options2)
    print(f'f1={f1}')
    print(f'f2={f2}')

    path1 = SSHPath('ssh://ubuntu@localhost:2222/.')
    path1.ls()
    path1.fs.ls('/')



"""
Logic for loading raw results from the MLops DAG root dir.

Used by ./aggregate.py
"""
import ubelt as ub
from kwutil import util_pattern
from kwutil import util_parallel
from geowatch.utils import util_dotdict
import parse
import json
from geowatch.mlops import smart_result_parser


def build_tables(root_dpath, dag, io_workers, eval_nodes,
                 cache_resolved_results):
    import pandas as pd
    from kwutil import util_progress

    io_workers = util_parallel.coerce_num_workers(io_workers)
    print(f'io_workers={io_workers}')

    # Hard coded nodes of interest to gather. Should abstract later.
    # The user-defined pipelines should be responsible for providing the
    # methods needed to parse their outputs.
    node_eval_infos = [
        {'name': 'bas_pxl_eval', 'out_key': 'eval_pxl_fpath'},
        {'name': 'sc_poly_eval', 'out_key': 'eval_fpath'},
        {'name': 'bas_poly_eval', 'out_key': 'eval_fpath'},
        {'name': 'sv_poly_eval', 'out_key': 'eval_fpath'},
    ]
    lut = ub.udict({info['name']: info for info in node_eval_infos})

    DEVFLAG = 1
    if DEVFLAG:
        for node_name in eval_nodes:
            node = dag.nodes[node_name]
            if node_name not in lut:
                if len(node.out_paths) == 1:
                    primary_out_key = list(node.out_paths)[0]
                elif getattr(node, 'primary_out_key', None) is not None:
                    primary_out_key = node.primary_out_key
                else:
                    raise Exception(ub.paragraph(
                        '''
                        evaluation nodes must have a single item in out_paths
                        or define a primary_out_key
                        '''))
                node_eval_infos.append({
                    'name': node.name,
                    'out_key': primary_out_key,
                })

    lut = ub.udict({info['name']: info for info in node_eval_infos})

    if eval_nodes is None:
        node_eval_infos_chosen = node_eval_infos
    else:
        try:
            node_eval_infos_chosen = list(lut.take(eval_nodes))
        except Exception as ex:
            from kwutil.util_exception import add_exception_note
            raise add_exception_note(ex, ub.paragraph(
                f'''
                Unknown evaluation node. Evaluation nodes need to be
                connected to a function that can parse their results.

                Requested evaluation nodes were: {eval_nodes}.
                But available nodes are {list(lut.keys())}.
                '''))

    from concurrent.futures import as_completed
    pman = util_progress.ProgressManager(backend='rich')
    # pman = util_progress.ProgressManager(backend='progiter')
    with pman:
        eval_type_to_results = {}

        eval_node_prog = pman.progiter(node_eval_infos_chosen, desc='Loading node results')

        for node_eval_info in eval_node_prog:
            node_name = node_eval_info['name']
            out_key = node_eval_info['out_key']

            if node_name not in dag.nodes:
                continue

            node = dag.nodes[node_name]
            out_node = node.outputs[out_key]

            fpaths = out_node_matching_fpaths(out_node)

            # Pattern match
            # node.template_out_paths[out_node.name]
            cols = {
                'index': [],
                'metrics': [],
                'requested_params': [],
                'resolved_params': [],
                'specified_params': [],
                'other': [],
                'fpath': [],
                # 'json_info': [],
            }

            executor = ub.Executor(mode='process', max_workers=io_workers)
            jobs = []
            submit_prog = pman.progiter(
                fpaths, desc=f'  * submit load jobs: {node_name}',
                transient=True)
            for fpath in submit_prog:
                job = executor.submit(load_result_worker, fpath, node_name,
                                      node=node, dag=dag,
                                      use_cache=cache_resolved_results)
                jobs.append(job)

            num_ignored = 0
            job_iter = as_completed(jobs)
            del jobs
            collect_prog = pman.progiter(
                job_iter, total=len(fpaths),
                desc=f'  * loading node results: {node_name}')
            for job in collect_prog:
                result = job.result()
                if result['requested_params'] or True:
                    assert set(result.keys()) == set(cols.keys())
                    for k, v in result.items():
                        cols[k].append(v)
                else:
                    num_ignored += 1

            if num_ignored:
                print(f'num_ignored = {ub.urepr(num_ignored, nl=1)}')

            results = {
                'fpath': pd.DataFrame(cols['fpath'], columns=['fpath']),
                'index': pd.DataFrame(cols['index']),
                'metrics': pd.DataFrame(cols['metrics']),
                'requested_params': pd.DataFrame(cols['requested_params'], dtype=object),  # prevents nones from being read as nan
                'specified_params': pd.DataFrame(cols['specified_params']),
                'resolved_params': pd.DataFrame(cols['resolved_params'], dtype=object),
                'other': pd.DataFrame(cols['other']),
            }
            # print(results['resolved_params']['resolved_params.sc_poly.smoothing'])
            eval_type_to_results[node_name] = results

    return eval_type_to_results


def load_result_worker(fpath, node_name, node=None, dag=None, use_cache=True):
    """
    Main driver for loading results

    Ignore:
        fpath = ub.Path('/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_testpipe/eval/flat/bas_poly_eval/bas_poly_eval_id_1ad531cc/poly_eval.json')
        node_name = 'bas_poly_eval'
    """
    import json
    import safer
    import rich
    from kwutil import util_json
    from kwutil.util_exception import add_exception_note
    fpath = ub.Path(fpath)

    resolved_json_fpath = fpath.parent / 'resolved_result_row_v012.json'

    if use_cache and resolved_json_fpath.exists():
        # Load the cached row data
        try:
            result = json.loads(resolved_json_fpath.read_text())
        except Exception as ex:
            raise add_exception_note(ex, f'Failed to read {resolved_json_fpath!r}')
    else:
        node_dpath = fpath.parent

        node_dpath = ub.Path(node_dpath)
        # Read the requested config
        job_config_fpath = node_dpath / 'job_config.json'
        if job_config_fpath.exists():
            try:
                job_config_text = job_config_fpath.read_text()
                _requested_params = json.loads(job_config_text)
            except Exception as ex:
                raise add_exception_note(ex, f'Failed to parse json job config {job_config_fpath}')
        else:
            _requested_params = {}

        requested_params = util_dotdict.DotDict(_requested_params).add_prefix('params')
        specified_params = {'specified.' + k: 1 for k in requested_params}

        # Read the resolved config
        # (Uses the DAG to trace the result lineage)
        try:
            flat = load_result_resolved(node_dpath, node=node, dag=dag)

            HACK_FOR_REGION_ID = True
            if HACK_FOR_REGION_ID:
                # Munge data to get the region ids we expect
                candidate_keys = list(flat.query_keys('region_ids'))
                region_ids = None
                for k in candidate_keys:
                    region_ids = flat[k]
                if region_ids is None:
                    print(ub.paragraph(
                        '''
                        Warning: no region ids available, some assumptions may
                        be violated.
                        '''))
                    region_ids = 'unknown'
                else:
                    IS_SMART = 0
                    if IS_SMART:
                        # Disable hack for smart region id names
                        import re
                        region_pat = re.compile(r'[A-Z][A-Za-z]*_[A-Z]\d\d\d')
                        region_ids = ','.join(list(region_pat.findall(region_ids)))

            resolved_params_keys = list(flat.query_keys('resolved_params'))
            metrics_keys = list(flat.query_keys('metrics'))
            resolved_params = flat & resolved_params_keys
            metrics = flat & metrics_keys

            other = flat - (resolved_params_keys + metrics_keys)

            index = {
                'node': node_name,
                'region_id': region_ids,
            }
            result = {
                'fpath': fpath,
                'index': index,
                'metrics': metrics,
                'requested_params': requested_params,
                'resolved_params': resolved_params,
                'specified_params': specified_params,
                'other': other,
            }

            # Cache this resolved row data
            result = util_json.ensure_json_serializable(result)
        except Exception as ex:
            rich.print(f'[red]Failed to load results for: {node_name}')
            rich.print(f'node_dpath={str(node_dpath)!r}')
            rich.print('ex = {}'.format(ub.urepr(ex, nl=1)))
            raise

        with safer.open(resolved_json_fpath, 'w') as file:
            json.dump(result, file, indent=4)

    HACK_FOR_BAD_SV_REQUEST = 1
    if HACK_FOR_BAD_SV_REQUEST:
        if node_name == 'sv_poly_eval':
            import pandas as pd
            if pd.isnull(result['requested_params'].get('params.sv_crop.crop_src_fpath', None)):
                result['requested_params'] = None

    return result


def new_process_context_parser(proc_item):
    """
    Load parameters out of data saved by a ProcessContext object
    """
    tracker_name_pat = util_pattern.MultiPattern.coerce({
        'geowatch.cli.run_tracker',
        'geowatch.cli.run_tracker',
    })
    heatmap_name_pat = util_pattern.MultiPattern.coerce({
        'geowatch.tasks.fusion.predict',
    })
    pxl_eval_pat = util_pattern.MultiPattern.coerce({
        'geowatch.tasks.fusion.evaluate',
    })
    proc_item = smart_result_parser._handle_process_item(proc_item)
    props = proc_item['properties']

    # Node-specific hacks
    params = props['config']
    if tracker_name_pat.match(props['name']):
        params.update(**json.loads(params.pop('track_kwargs', '{}')))
    elif heatmap_name_pat.match(props['name']):
        params.pop('datamodule_defaults', None)
    elif pxl_eval_pat.match(props['name']):
        from geowatch.tasks.fusion import evaluate
        # We can resolve the params to a dictionary in this instance
        if isinstance(params, list) or 'true_dataset' not in params:
            args = props['args']
            params = evaluate.SegmentationEvalConfig().load(cmdline=args).to_dict()

    resources = smart_result_parser.parse_resource_item(proc_item, add_prefix=False)

    output = {
        # TODO: better name for this
        'context': {
            'task': props['name'],
            'uuid': props.get('uuid', None),
            'start_timestamp': props.get('start_timestamp', None),
            'stop_timestamp': props.get('stop_timestamp', props.get('end_timestamp', None)),
        },
        'resolved_params': params,
        'resources': resources,
        'machine': props.get('machine', {}),
    }
    return output


def load_result_resolved(node_dpath, node=None, dag=None):
    """
    Recurse through the DAG filesytem structure and load resolved
    configurations from each step.

    Args:
        node_dpath (str | PathLike):
            the path to the evaluation node directory. The specific type of
            evaluation node must have a known (currently hard-coded) condition
            in this function that knows how to parse it.

        node (None | ProcessNode):
            new experimental way to allow users to specify how results should
            be loaded. The node should have a "load_result" function that
            accepts node_dpath as a single argument and then returns a flat
            resolved dotdict of hyperparams, metrics, and context.

        dag (None | Pipeline):
            Used to lookup loading functions for predecessor nodes.

    Returns:
        Dict - flat_resolved - a flat dot-dictionary with resolved params

    TODO:
        Some mechanism to let the user specify how to parse an evaluation node
        of a given type.

    Ignore:
        >>> # To diagnose issues, construct a path to an evaluation node to get the
        >>> # relevant project-specific entrypoint data.
        >>> # TODO: need a demo pipeline that we can test for robustness here.
        >>> from geowatch.mlops.aggregate_loader import *  # NOQA
        >>> from geowatch.mlops.aggregate_loader import load_result_resolved
        >>> import geowatch
        >>> import rich
        >>> expt_dpath = geowatch.find_dvc_dpath(tags='phase3_expt')
        >>> # choose the location mlops schedule dumped results to
        >>> mlops_dpath = expt_dpath / '_preeval20_bas_grid'
        >>> # Search and pick a poly eval node, the specific path
        >>> # will depend on the pipeline structure, which may be revised
        >>> # in the future. At the start of SMART phase3, we keep all
        >>> # eval nodes grouped in eval/flat, so enumerate those
        >>> node_type_dpaths = list(mlops_dpath.glob('eval/flat/*'))
        >>> node_type_dpaths += list(mlops_dpath.glob('pred/flat/*'))
        >>> # For each eval node_type, choose a node in it.
        >>> for node_type_dpath in node_type_dpaths:
        >>>     for node_dpath in node_type_dpath.ls():
        >>>         if len(node_dpath.ls()) > 2:
        >>>             print(f'Found node_dpath={node_dpath}')
        >>>             break
        >>>     print(f'node_dpath={node_dpath}')
        >>>     flat_resolved = load_result_resolved(node_dpath)
        >>>     rich.print(f'flat_resolved = {ub.urepr(flat_resolved, nl=1)}')

    Ignore:
        ## OR If you know the node_type of node you want
        from geowatch.mlops.aggregate_loader import *  # NOQA
        import geowatch
        import rich
        expt_dpath = geowatch.find_dvc_dpath(tags='phase3_expt')
        mlops_dpath = expt_dpath / '_preeval20_bas_grid'

        node_group, node_type = 'eval', 'bas_poly_eval'
        node_group, node_type = 'pred', 'sc_pxl'
        node_group, node_type = 'pred', 'bas_pxl'
        node_group, node_type = 'eval', 'bas_pxl_eval'

        for node_dpath in mlops_dpath.glob(f'{node_group}/flat/{node_type}/*'):
            if len(node_dpath.ls()) > 2:
                print(f'Found node_dpath={node_dpath}')
                break

        flat_resolved = load_result_resolved(node_dpath)
        rich.print(f'flat_resolved = {ub.urepr(flat_resolved, nl=1)}')

    Ignore:
        from geowatch.mlops.aggregate_loader import *  # NOQA
        node_dpath = ub.Path('/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_testpipe/eval/flat/bas_poly_eval/bas_poly_eval_id_1ad531cc')
        node_dpath = ub.Path('/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_testpipe/eval/flat/bas_pxl_eval/bas_pxl_eval_id_6028edfe/')
        node_dpath = ub.Path('/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_timekernel_test_drop4/eval/flat/bas_pxl_eval/bas_pxl_eval_id_5d38c6b3')
        got = load_result_resolved(node_dpath)
    """
    # from geowatch.utils.util_dotdict import explore_nested_dict
    node_dpath = ub.Path(node_dpath)
    node_type_dpath = node_dpath.parent
    node_type = node_type_dpath.name

    if dag is not None:
        if node is None:
            try:
                node = dag.nodes[node_type]
            except KeyError:
                print(f'node_dpath = {ub.urepr(node_dpath, nl=1)}')
                print(f'node_type = {ub.urepr(node_type, nl=1)}')
                print(f'dag.nodes = {ub.urepr(dag.nodes, nl=1)}')
                raise

    if node is not None and hasattr(node, 'load_result'):
        flat_resolved = node.load_result(node_dpath)
        if flat_resolved is None:
            raise AssertionError('node.load_result should have returned a dict')

    elif node_type in {'sc_pxl', 'bas_pxl'}:
        pat = util_pattern.Pattern.coerce(node_dpath / 'pred.kwcoco.*')
        found = list(pat.paths())
        if len(found) == 0:
            raise FileNotFoundError(f'Unable to find expected kwcoco file in {node_type} node_dpath: {node_dpath}')
        fpath = found[0]
        bas_pxl_info = smart_result_parser.parse_json_header(fpath)
        proc_item = smart_result_parser.find_pred_pxl_item(bas_pxl_info)
        nest_resolved = new_process_context_parser(proc_item)
        flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
        flat_resolved = flat_resolved.insert_prefix(node_type, index=1)

        # Record the train-time parameters
        fit_node_type = node_type + '_fit'
        extra = proc_item['properties']['extra']
        fit_config = extra['fit_config']
        if 'data' not in fit_config:
            raise Exception(ub.paragraph(
                f'''
                A kwcoco has an old fit-config that did not contain all
                train-time params. To fix this run for a single file run:
                ``python -m geowatch.cli.experimental.fixup_predict_kwcoco_metadata {fpath}``
                ''') +
                '\n\n' +
                ub.paragraph(
                    '''
                    For more details see:
                    ``python -m geowatch.cli.experimental.fixup_predict_kwcoco_metadata --help``
                    '''))

        fit_nested = {
            'context': {'task': 'geowatch.tasks.fusion.fit'},
            'resolved_params': fit_config,
            'resources': {},
            'machine': {},
        }
        flat_fit_resolved = util_dotdict.DotDict.from_nested(fit_nested)
        flat_fit_resolved = flat_fit_resolved.insert_prefix(fit_node_type, index=1)
        flat_resolved |= flat_fit_resolved

    elif node_type in {'bas_poly', 'sc_poly'}:
        pat = util_pattern.Pattern.coerce(node_dpath / 'poly.kwcoco.*')
        fpath = list(pat.paths())[0]
        try:
            bas_poly_info = smart_result_parser.parse_json_header(fpath)
        except Exception:
            # There are some cases where the kwcoco file was clobbered,
            # but we can work around by using a manifest file.
            pat = util_pattern.Pattern.coerce(node_dpath / '*_manifest.json')
            fpath = list(pat.paths())[0]
            bas_poly_info = smart_result_parser.parse_json_header(fpath)
            raise

        proc_item = smart_result_parser.find_track_item(bas_poly_info)
        nest_resolved = new_process_context_parser(proc_item)
        flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
        flat_resolved = flat_resolved.insert_prefix(node_type, index=1)

    elif node_type in {'bas_poly_eval', 'sc_poly_eval', 'sv_poly_eval'}:
        fpath = node_dpath / 'poly_eval.json'
        iarpa_result = smart_result_parser.load_iarpa_evaluation(fpath)
        proc_item = smart_result_parser.find_metrics_framework_item(
            iarpa_result['iarpa_json']['info'])
        nest_resolved = new_process_context_parser(proc_item)
        nest_resolved['metrics'] = iarpa_result['metrics']

        flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
        flat_resolved['context.region_ids'] = iarpa_result['iarpa_json']['region_ids']
        flat_resolved = flat_resolved.insert_prefix(node_type, 1)

    elif node_type in {'bas_pxl_eval', 'sc_pxl_eval'}:
        fpath = node_dpath / 'pxl_eval.json'
        info = smart_result_parser.load_pxl_eval(fpath, with_param_types=False)
        metrics = info['metrics']

        proc_item = smart_result_parser.find_pxl_eval_item(
            info['json_info']['meta']['info'])

        nest_resolved = new_process_context_parser(proc_item)
        # Hack for region ids
        nest_resolved['context']['region_ids'] = ub.Path(nest_resolved['resolved_params']['true_dataset']).name.split('.')[0]
        nest_resolved['metrics'] = metrics

        flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
        flat_resolved = flat_resolved.insert_prefix(node_type, 1)

    elif node_type in {'valicrop', 'sv_crop'}:

        # TODO: parse resolved params
        # nest_resolved = {}
        # flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
        # flat_resolved = flat_resolved.insert_prefix(node_type, index=1)
        node_process_name = 'coco_align'
        fpath = node_dpath / 'sv_crop.kwcoco.zip'
        if not fpath.exists():
            fpath = node_dpath / 'valicrop.kwcoco.zip'
        flat_resolved = _generalized_process_flat_resolved(fpath, node_process_name, node_type)

    elif node_type in {'sc_crop'}:
        # TODO: parse resolved params
        node_process_name = 'coco_align'
        fpath = node_dpath / 'sitecrop.kwcoco.zip'
        flat_resolved = _generalized_process_flat_resolved(fpath, node_process_name, node_type)

    elif node_type in {'cluster_sites'}:
        # TODO: write out resolved params in cluster sites
        flat_resolved = {}

    elif node_type in {'sv_dino_boxes'}:
        node_process_name = 'box.predict'
        fpath = node_dpath / 'pred_boxes.kwcoco.zip'
        flat_resolved = _generalized_process_flat_resolved(fpath, node_process_name, node_type)

    elif node_type in {'sv_dino_filter'}:
        node_process_name = 'geowatch.tasks.dino_detector.building_validator'
        fpath = node_dpath / 'out_site_manifest.json'
        flat_resolved = _generalized_process_flat_resolved(fpath, node_process_name, node_type)
    # TODO: it would be nice if just declaring a node gave us this information.
    elif node_type in {'sv_depth_score'}:
        from geowatch.mlops import smart_pipeline
        fpath = node_dpath / smart_pipeline.SV_DepthPredict.out_paths['out_kwcoco']
        node_process_name = 'geowatch.tasks.depth_pcd.score_tracks'
        flat_resolved = _generalized_process_flat_resolved(fpath, node_process_name, node_type)
    elif node_type in {'sv_depth_filter'}:
        from geowatch.mlops import smart_pipeline
        fpath = node_dpath / smart_pipeline.SV_DepthFilter.out_paths['output_site_manifest_fpath']
        node_process_name = 'geowatch.tasks.depth_pcd.filter_tracks'
        flat_resolved = _generalized_process_flat_resolved(fpath, node_process_name, node_type)
    else:
        raise NotImplementedError(ub.paragraph(
            f'''
            Attempted to load a result for {node_type} in {node_dpath}.
            But was unable to determine how to do so.
            In your pipeline class you a method ``def load_result(self,
            node_dpath):`` which returns a flat dot-dictionary of params and
            results from the node.
            '''))

    # Determine if this node has any predecessor computations and load results
    # from those as well to have a flat and complete picture of the process
    # lineage for this node.
    predecessor_dpath = node_dpath / '.pred'
    for predecessor_node_type_dpath in predecessor_dpath.glob('*'):
        # predecessor_node_type = predecessor_node_type_dpath.name
        for predecessor_node_dpath in predecessor_node_type_dpath.glob('*'):
            if predecessor_node_dpath.exists():
                predecessor_flat_resolved = load_result_resolved(predecessor_node_dpath, dag=dag)
                flat_resolved |= predecessor_flat_resolved

    return flat_resolved


def _generalized_process_flat_resolved(fpath, node_process_name, node_type):
    """
    Parses info out of json files where we conform to a pattern that the top
    level json object has an "info" section as list of ProcessContext info
    objects. We grab the context of a metric-less node there. We do need to
    know the filepath of this json file and what the name used in process
    context was.
    """
    info = smart_result_parser.parse_json_header(fpath)
    proc_item = smart_result_parser.find_info_items(info, {'process'}, {node_process_name})
    items = list(proc_item)
    assert len(items) == 1
    proc_item = items[0]
    nest_resolved = new_process_context_parser(proc_item)
    flat_resolved = util_dotdict.DotDict.from_nested(nest_resolved)
    flat_resolved = flat_resolved.insert_prefix(node_type, index=1)
    return flat_resolved


def out_node_matching_fpaths(out_node):
    out_template = out_node.template_value
    parser = parse.Parser(str(out_template))
    patterns = {n: '*' for n in parser.named_fields}
    pat = out_template.format(**patterns)
    mpat = util_pattern.Pattern.coerce(pat)
    fpaths = list(mpat.paths())
    return fpaths


if 1:
    import numpy as np
    if np.bool_ is not bool:
        # Hack for a ubelt issue
        @ub.hash_data.register(np.bool_)
        def _hashnp_bool(data):
            from ubelt.util_hash import _int_to_bytes
            # warnings.warn('Hashing ints is slow, numpy is preferred')
            hashable = _int_to_bytes(bool(data))
            # hashable = data.to_bytes(8, byteorder='big')
            prefix = b'INT'
            return prefix, hashable



# ---
#
# Initial code from https://github.com/milesial/Pytorch-UNet
#
# ---
from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
# from pytorch_memlab import profile, profile_every


class UNetBackbone(nn.Module):
    def __init__(self, n_channels, n_hidden_base, no_padding=False):
        super(UNetBackbone, self).__init__()
        self.no_padding = no_padding
        self.inc = InConv(n_channels, n_hidden_base, no_padding)
        self.down1 = Down(n_hidden_base, n_hidden_base * 2, no_padding)
        self.down2 = Down(n_hidden_base * 2, n_hidden_base * 4, no_padding)
        self.down3 = Down(n_hidden_base * 4, n_hidden_base * 8, no_padding)
        self.down4 = Down(n_hidden_base * 8, n_hidden_base * 16, no_padding)
        # self.up1 = Up(n_hidden_base*24, n_hidden_base*8, no_padding)
        # self.up2 = Up(n_hidden_base*12, n_hidden_base*4, no_padding)
        # self.up3 = Up(n_hidden_base*6, n_hidden_base*2, no_padding)
        # self.up4 = Up(n_hidden_base*3, n_hidden_base, no_padding)

        self.up1 = Up(n_hidden_base * 16, n_hidden_base * 8, n_hidden_base * 8, no_padding)
        self.up2 = Up(n_hidden_base * 8, n_hidden_base * 4, n_hidden_base * 4, no_padding)
        self.up3 = Up(n_hidden_base * 4, n_hidden_base * 2, n_hidden_base * 2, no_padding)
        self.up4 = Up(n_hidden_base * 2, n_hidden_base, n_hidden_base, no_padding)

    def forward(self, x):
        x0 = self.inc.forward(x)
        x1 = self.down1.forward(x0)
        x2 = self.down2.forward(x1)
        x3 = self.down3.forward(x2)
        y4 = self.down4.forward(x3)
        y3 = self.up1.forward(y4, x3)
        y2 = self.up2.forward(y3, x2)
        y1 = self.up3.forward(y2, x1)
        y0 = self.up4.forward(y1, x0)

        result = OrderedDict()
        result["out"] = y0

        return result


class DoubleConv(nn.Module):
    """(conv => BN => ReLU) * 2"""

    def __init__(self, in_ch, out_ch, no_padding):
        super(DoubleConv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch, out_ch, 3, padding=0 if no_padding else 1, bias=True),
            nn.BatchNorm2d(out_ch),
            nn.ELU(),
            nn.Conv2d(out_ch, out_ch, 3, padding=0 if no_padding else 1, bias=True),
            nn.BatchNorm2d(out_ch),
            nn.ELU()
        )

    # @profile
    def forward(self, x):
        x = self.conv(x)
        return x


class InConv(nn.Module):
    def __init__(self, in_ch, out_ch, no_padding):
        super(InConv, self).__init__()
        self.conv = DoubleConv(in_ch, out_ch, no_padding)

    def forward(self, x):
        x = self.conv.forward(x)
        return x


class Down(nn.Module):
    def __init__(self, in_ch, out_ch, no_padding):
        super(Down, self).__init__()
        self.mpconv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_ch, out_ch, no_padding)
        )

    def forward(self, x):
        x = self.mpconv(x)
        return x


class Up(nn.Module):
    def __init__(self, in_ch_1, in_ch_2, out_ch, no_padding):
        super(Up, self).__init__()
        # self.upsample = nn.ConvTranspose2d(in_ch_1, in_ch_2, 3, stride=2, padding=1)
        self.conv = DoubleConv(in_ch_1 + in_ch_2, out_ch, no_padding)

    def forward(self, x1, x2):
        # print("--- Up ---")
        x1 = F.interpolate(x1, scale_factor=2, mode='bilinear', align_corners=False)
        # upsample_shape = [x1.shape[0], x2.shape[1], 2*x1.shape[2], 2*x1.shape[3]]
        # x1 = self.upsample(x1, output_size=upsample_shape)

        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])

        # for padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd

        x = torch.cat([x2, x1], dim=1)
        x = self.conv.forward(x)
        return x



import ubelt as ub
import os
import pytest


def test_top_level_cli_help_message():
    verbose = 0
    exe = 'python -m geowatch'

    env = dict(os.environ)
    env['WATCH_PREIMPORT'] = '0'

    kwargs = {
        'verbose': verbose,
        'env': env,
    }

    info = ub.cmd(f'{exe} --help', **kwargs)
    assert info['ret'] == 0


def test_subcommand_cli_help_message():
    if not int(os.environ.get('SLOW_TESTS', '0')):
        pytest.skip('skip slower test')

    verbose = 0
    exe = 'python -m geowatch'

    env = dict(os.environ)
    env['WATCH_PREIMPORT'] = '0'

    kwargs = {
        'verbose': verbose,
        'env': env,
    }

    info = ub.cmd(f'{exe} align --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} dvcdir --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} find_dvc --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} stats --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} add_fields --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} model_stats --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} reproject --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} average_features --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} time_combine --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} iarpa_eval --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} spectra --help', **kwargs)
    assert info['ret'] == 0

    info = ub.cmd(f'{exe} visualize --help', **kwargs)
    assert info['ret'] == 0




import torch

class MyMWE(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.stem1 = torch.nn.Linear(2, 3)
        self.stem2 = torch.nn.Linear(2, 3)
        self.body = torch.nn.Linear(3, 3)
        self.head1 = torch.nn.Linear(3, 5)
        self.head2 = torch.nn.Linear(3, 7)

    def forward(self, batch):
        batch_outputs = []
        for item in batch:
            if 'domain1' in item:
                feat = self.stem1(item['domain1'])
            elif 'domain2' in item:
                feat = self.stem2(item['domain2'])
            else:
                raise ValueError

            hidden = self.body(feat)
            logits1 = self.head1(hidden)
            logits2 = self.head2(hidden)
            output = {
                'logits1': logits1,
                'logits2': logits2,
            }
            batch_outputs.append(output)
        return output

def main():
    from torchview import draw_graph
    model = MyMWE()

    batch = [
        {'domain1': torch.Tensor([1, 2])},
        {'domain2': torch.Tensor([1, 2])},
        {'domain1': torch.Tensor([1, 2])},
    ]

    # Verify a normal forward pass works
    output = model(batch)

    # Check if draw_graph works
    model_graph = draw_graph(
        model,
        input_data=batch,
        expand_nested=True,
        hide_inner_tensors=True,
        device='meta', depth=np.inf)
    model_graph.visual_graph.view()
    model_graph.visual_graph.render(format='png')



from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_datamodule import KWCocoVideoDataModuleConfig  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.kwcoco_datamodule import KWCocoVideoDataModule  # NOQA



#!/usr/bin/env python3
import os
import scriptconfig as scfg
import ubelt as ub


class DetectronPredictCLI(scfg.DataConfig):
    # TODO: scriptconfig Value classes should have tags for mlops pipelines
    # Something like tags ⊆ {in_path,out_path, algo_param, perf_param, primary, primary_in_path, primary_out_path}
    checkpoint_fpath = scfg.Value(None, help='path to the weights')
    model_fpath = scfg.Value(None, help='path to a model file: todo: bundle with weights')

    src_fpath = scfg.Value(None, help='input kwcoco file')
    dst_fpath = scfg.Value(None, help='output kwcoco file')
    write_heatmap = scfg.Value(False, help='if True, also write masks as heatmaps')
    nms_thresh = scfg.Value(0.0, help='nonmax supression threshold')
    workers = 4
    device = scfg.Value('cuda:0', help='device to predict on')

    base = scfg.Value('auto', help=ub.paragraph(
        '''
        Specify the detectron config that corresponds to the model.
        If "auto" attempts to introspect from the checkpoint fpath, which is
        currently only possible if the model was written as a sidecar file.

        Could be something like:
            * COCO-Detection/faster_rcnn_R_50_FPN_3x.yaml
            * COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml
            * new_baselines/mask_rcnn_regnety_4gf_dds_FPN_400ep_LSJ.py
        '''))

    cfg = scfg.Value(ub.codeblock(
        '''
        '''),
        help=ub.paragraph(
            '''
            Overlaid config on top of whatever base config path is specified.
            Typically this is not needed at test time, but it provided just in
            case.
            '''))

    @classmethod
    def main(cls, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> cmdline = 0
            >>> kwargs = dict()
            >>> cls = DetectronPredictCLI
            >>> cls.main(cmdline=cmdline, **kwargs)
        """
        import rich
        from rich.markup import escape
        config = cls.cli(cmdline=cmdline, data=kwargs, strict=True)
        rich.print('config = ' + escape(ub.urepr(config, nl=1)))
        detectron_predict(config)

__cli__ = DetectronPredictCLI


class OldStyleConfigBackend:
    """
    Handle configuration for old-style detectron2 configs
    """
    def __init__(backend, cfg, config):
        backend.cfg = cfg
        backend.config = config
        backend.oldstyle_predictor = None
        backend.model = None

    def load_model(backend):
        # NEED to call detectron2 more efficiently
        from detectron2.engine import DefaultPredictor
        backend.oldstyle_predictor = DefaultPredictor(backend.cfg)
        backend.model = backend.oldstyle_predictor.model

    def preprocess(backend, im_hwc):
        image = backend.oldstyle_predictor.aug.get_transform(im_hwc).apply_image(im_hwc)
        return image

    @classmethod
    def from_config(cls, config):
        # FIXME: remove hard coding
        from detectron2.config import get_cfg
        from detectron2 import model_zoo
        from detectron2.config import CfgNode
        import kwutil
        cfg = get_cfg()
        base_cfg = model_zoo.get_config_file(config.base)
        cfg.merge_from_file(base_cfg)

        # cfg.DATASETS.TRAIN = (dataset_infos['train']['name'],)
        # cfg.DATASETS.TEST = (dataset_infos['vali']['name'],)
        # cfg.DATASETS.TEST = ()
        # cfg.DATALOADER.NUM_WORKERS = 2
        # cfg.MODEL.WEIGHTS = model_zoo.get_checkpoint_url('COCO-InstanceSegmentation/mask_rcnn_R_50_FPN_3x.yaml')  # Let training initialize from model zoo
        # cfg.SOLVER.IMS_PER_BATCH = 2   # This is the real 'batch size' commonly known to deep learning people
        # cfg.SOLVER.BASE_LR = 0.00025   # pick a good LR
        # cfg.SOLVER.MAX_ITER = 120_000  # 300 iterations seems good enough for this toy dataset; you will need to train longer for a practical dataset
        # cfg.SOLVER.STEPS = []          # do not decay learning rate
        # cfg.MODEL.ROI_HEADS.BATCH_SIZE_PER_IMAGE = 128   # The 'RoIHead batch size'. 128 is faster, and good enough for this toy dataset (default: 512)
        # cfg.MODEL.ROI_HEADS.NUM_CLASSES = 46  # only has one class (ballon). (see https://detectron2.readthedocs.io/tutorials/datasets.html#update-the-config-for-new-datasets)
        # cfg.MODEL.ROI_HEADS.NUM_CLASSES = 3  # only has one class (ballon). (see https://detectron2.readthedocs.io/tutorials/datasets.html#update-the-config-for-new-datasets)
        # NOTE: this config means the number of classes, but a few popular unofficial tutorials incorrect uses num_classes+1 here.
        # cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = 0.1  # set threshold for this model

        ckpt_path = config.checkpoint_fpath
        if ckpt_path is not None or ckpt_path != 'noop':
            if not os.path.exists(ckpt_path) and ckpt_path.endswith('yaml'):
                # Checkpoint is likely given as a model zoo resource
                ckpt_path = model_zoo.get_checkpoint_url(ckpt_path)
            cfg.MODEL.WEIGHTS = os.fspath(ckpt_path)

        backend = cls(cfg, config)

        cfg_final_layer = kwutil.Yaml.coerce(config.cfg, backend='pyyaml')
        if cfg_final_layer:
            cfg2 = CfgNode(cfg_final_layer)
            print(ub.urepr(cfg2, nl=-1))
            cfg.merge_from_other_cfg(cfg2)
        print(ub.urepr(cfg, nl=-1))

        backend.device = backend.cfg.MODEL.DEVICE
        return backend


class NewStyleConfigBackend:
    """
    Handle configuration for new-style "lazy" detectron2 configs
    """
    def __init__(backend, cfg, config):
        backend.cfg = cfg
        backend.config = config
        backend.model = None
        backend.device = None

    def preprocess(backend, im_hwc):
        return im_hwc

    def load_model(backend):
        from detectron2.config import instantiate
        from detectron2.checkpoint import DetectionCheckpointer
        import torch
        model = instantiate(backend.cfg.model)
        checkpointer = DetectionCheckpointer(model)
        info = checkpointer.load(backend.config.checkpoint_fpath)  # NOQA
        model.eval()
        backend.device = torch.device(backend.config.device)
        model.to(backend.device)
        backend.model = model

    @classmethod
    def from_train_dpath(cls, train_dpath, config):
        import kwutil
        import kwcoco
        yaml_files = list(train_dpath.glob('*.yaml'))
        if len(yaml_files) == 0:
            raise Exception('no yaml files')
        elif len(yaml_files) > 1:
            raise NotImplementedError('need to find the right yaml')
        config_path = yaml_files[0]
        candidate_pkl = config_path.augment(tail='.pkl')
        # If a pickle file exists as a yaml sidecar we likely need to use that
        # instead as the yaml likely wont fully resolve.
        if candidate_pkl.exists():
            import pickle
            with open(candidate_pkl, 'rb') as fp:
                cfg = pickle.load(fp)
        else:
            cfg = kwutil.Yaml.load(config_path)
        backend = cls(cfg, config)

        cat_fpath = train_dpath / 'categories.json'

        # TODO: this should really be handled on a head-by-head basis
        if 'classes' in backend.cfg:
            classes = kwcoco.CategoryTree.coerce(backend.cfg['classes'])
        else:
            if cat_fpath.exists():
                print('found categories fpath')
                categories = kwutil.Json.load(cat_fpath)
                classes = kwcoco.CategoryTree.from_coco(categories)
                classes = list(classes)
            else:
                raise NotImplementedError('categories.json not found, no way to procede without classes')
        backend.classes = classes

        # Override config values
        cfg_final_layer = kwutil.Yaml.coerce(config.cfg, backend='pyyaml')
        if cfg_final_layer:
            walker = ub.IndexableWalker(cfg)
            to_set_walker = ub.IndexableWalker(cfg_final_layer)
            for p, v in to_set_walker:
                if not isinstance(v, dict):
                    walker[p] = v
        return backend


class Detectron2Predictor:

    def __init__(predictor, config):
        predictor.config = config
        predictor.backend = None
        predictor.dset = None

    def prepare_config_backend(predictor):
        config = predictor.config
        if config.base == 'auto':
            # Attempt to find the model configuration file based on the
            # checkpoint path.
            checkpoint_fpath = ub.Path(config.checkpoint_fpath)
            train_dpath = checkpoint_fpath.parent
            backend = NewStyleConfigBackend.from_train_dpath(train_dpath, config)
        else:
            backend = OldStyleConfigBackend.from_config(config)
        predictor.backend = backend
        predictor.backend.load_model()

    @property
    def classes(self):
        assert self.backend is not None
        return self.backend.classes

    def prepare_dataset(predictor):
        import kwcoco
        dset = kwcoco.CocoDataset(predictor.config.src_fpath)
        keep_annots = 0
        if not keep_annots:
            dset.clear_annotations()
        dset.reroot(absolute=True)

        # import geowatch
        from geowatch.tasks.fusion.datamodules.kwcoco_dataset import KWCocoVideoDataset
        dataset = KWCocoVideoDataset(
            dset,
            window_space_dims='full',
            channels='blue|green|red',
            time_dims=1,
            mode='test',
            # todo: enhance reduce item size to remove most information, but do
            # keep the image id.
            # reduce_item_size=True,
        )
        # batch_item = dataset[0]
        dset.reroot(absolute=True)
        bundle_dpath = ub.Path(predictor.config.dst_fpath).parent.ensuredir()
        dset.fpath = predictor.config.dst_fpath

        predictor.dset = dset
        predictor.dataset = dataset
        predictor.bundle_dpath = bundle_dpath

    def prepare_sticher(predictor):
        """
        """
        config = predictor.config
        if config.write_heatmap:
            from geowatch.tasks.fusion.coco_stitcher import CocoStitchingManager
            from kwutil import util_parallel
            writer_queue = util_parallel.BlockingJobQueue(
                mode='thread',
                # mode='serial',
                max_workers=2,
            )
            stitcher_common_kw = dict(
                stiching_space='image',
                device='numpy',
                write_probs=True,
                expected_minmax=(0, 1),
                writer_queue=writer_queue,
                assets_dname='_assets',
            )
            stitcher = CocoStitchingManager(
                predictor.dset,
                chan_code='salient',
                short_code='salient',
                num_bands=1,
                **stitcher_common_kw,
            )
        else:
            stitcher = None
            writer_queue = None
        predictor.stitcher = stitcher
        predictor.writer_queue = writer_queue

    def run_prediction(predictor):
        """
        batch prediction on the kwcoco file
        """
        # import kwimage
        # import kwarray
        import torch
        # import einops
        import numpy as np
        import rich
        # TODO: could be much more efficient
        # torch_impl = kwarray.ArrayAPI.coerce('torch')()
        # config = predictor.config

        # FIXME: We need a method to know what classes the detectron2 model was
        # trained with.
        # predictor.dset.object_categories()
        dset = predictor.dset
        print(f'dset={dset}')

        loader = predictor.dataset.make_loader(
            batch_size=1,
            num_workers=4,  # config.workers
        )
        # images = dset.images()
        bundle_dpath = predictor.bundle_dpath
        rich.print(f'Pred Dpath: [link={bundle_dpath}]{bundle_dpath}[/link]')

        batch_iter = ub.ProgIter(loader, total=len(loader), desc='predict')
        with torch.no_grad():  # https://github.com/sphinx-doc/sphinx/issues/4258
            for batch in batch_iter:
                assert len(batch) == 1
                batch_item = batch[0]
                image_id = batch_item['target']['gids'][0]
                im_chw = batch_item['frames'][0]['modes']['blue|green|red']

                dets = predictor.predict_image(im_chw)

                for ann in dets.to_coco(style='new'):
                    ann['image_id'] = image_id
                    catname = ann.pop('category_name')
                    ann['category_id'] = dset.ensure_category(catname)
                    ann['role'] = 'prediction'
                    dset.add_annotation(**ann)

                if predictor.stitcher is not None:
                    frame_info = batch_item['frames'][0]
                    output_image_dsize = frame_info['output_image_dsize']
                    output_space_slice = frame_info['output_space_slice']
                    scale_outspace_from_vid = frame_info['scale_outspace_from_vid']

                    sorted_dets = dets.take(dets.scores.argsort())
                    probs = np.zeros(output_image_dsize[::-1], dtype=np.float32)
                    for sseg, score in zip(sorted_dets.data['segmentations'], sorted_dets.scores):
                        sseg.data.fill(probs, value=float(score), assert_inplace=True)

                    predictor.stitcher.accumulate_image(
                        image_id, output_space_slice, probs,
                        asset_dsize=output_image_dsize,
                        scale_asset_from_stitchspace=scale_outspace_from_vid,
                        # weights=output_weights,
                        # downweight_edges=downweight_edges,
                    )
                    # hack / fixme: this is ok, when batches correspond with
                    # images but not if we start to window.
                    predictor.stitcher.submit_finalize_image(image_id)

        if predictor.stitcher is not None:
            predictor.writer_queue.wait_until_finished()  # hack to avoid race condition
            # Prediction is completed, finalize all remaining images.
            print(f"Finalizing stitcher for {predictor.stitcher}")
            for gid in predictor.stitcher.managed_image_ids():
                predictor.stitcher.submit_finalize_image(gid)
            predictor.writer_queue.wait_until_finished()

    def predict_image(predictor, im_chw):
        import torch
        import kwimage
        import kwarray
        import einops

        # TODO: ensure references are not costly to access
        # classes = COCO_CLASSES  # TODO: grab from config
        classes = predictor.classes
        config = predictor.config
        torch_impl = kwarray.ArrayAPI.coerce('torch')()

        im_hwc = einops.rearrange(im_chw, 'c h w -> h w c').numpy()

        # Preprocess the inputs
        height, width = im_chw.shape[1:3]
        image = predictor.backend.preprocess(im_hwc)
        image = torch.as_tensor(image.astype("float32").transpose(2, 0, 1))
        image = image.to(predictor.backend.device)

        inputs = {"image": image, "height": height, "width": width}
        outputs = predictor.backend.model([inputs])[0]

        instances = outputs['instances']
        if len(instances):
            boxes = instances.pred_boxes
            scores = instances.scores
            boxes = kwimage.Boxes(boxes.tensor, format='ltrb').numpy()
            scores = torch_impl.numpy(instances.scores)
            pred_class_indexes = torch_impl.numpy(instances.pred_classes)

            detkw = {
                'boxes': boxes,
                'scores': scores,
                'class_idxs': pred_class_indexes,
                'classes': classes,
            }
            if hasattr(instances, 'pred_masks'):
                pred_masks = torch_impl.numpy(instances.pred_masks)
                segmentations = []
                for cmask in pred_masks:
                    mask = kwimage.Mask.coerce(cmask)
                    poly = mask.to_multi_polygon()
                    segmentations.append(poly)
                detkw['segmentations'] = kwimage.SegmentationList(segmentations)
            else:
                # detkw['segmentations'] = None
                ...

            dets = kwimage.Detections(**detkw)
            if config.nms_thresh and config.nms_thresh > 0:
                dets = dets.non_max_supress(thresh=config.nms_thresh)
        else:
            dets = kwimage.Detections.random(0)
            dets.data['segmentations'] = kwimage.SegmentationList([])
        return dets


def detectron_predict(config):
    import geowatch_tpl
    detectron2 = geowatch_tpl.import_submodule('detectron2')  # NOQA
    import kwutil
    import rich

    proc_context = kwutil.ProcessContext(
        name='geowatch.tasks.detectron2.predict',
        config=kwutil.Json.ensure_serializable(dict(config)),
        track_emissions=True,
    )
    proc_context.start()
    print(f'proc_context.obj = {ub.urepr(proc_context.obj, nl=3)}')

    predictor = Detectron2Predictor(config)
    predictor.prepare_dataset()
    predictor.prepare_config_backend()
    predictor.prepare_sticher()
    predictor.run_prediction()

    # Write final results
    predictor.dset.dataset.setdefault('info', [])
    proc_context.stop()
    print(f'proc_context.obj = {ub.urepr(proc_context.obj, nl=3)}')
    predictor.dset.dataset['info'].append(proc_context.obj)
    predictor.dset.dump()
    bundle_dpath = predictor.bundle_dpath
    rich.print(f'Wrote to: [link={bundle_dpath}]{bundle_dpath}[/link]')


COCO_CLASSES = ['__background__', 'person', 'bicycle', 'car', 'motorcycle',
                'airplane', 'bus', 'train', 'truck', 'boat', 'traffic light',
                'fire hydrant', 'stop sign', 'parking meter', 'bench', 'bird',
                'cat', 'dog', 'horse', 'sheep', 'cow', 'elephant', 'bear',
                'zebra', 'giraffe', 'backpack', 'umbrella', 'handbag', 'tie',
                'suitcase', 'frisbee', 'skis', 'snowboard', 'sports ball',
                'kite', 'baseball bat', 'baseball glove', 'skateboard',
                'surfboard', 'tennis racket', 'bottle', 'wine glass', 'cup',
                'fork', 'knife', 'spoon', 'bowl', 'banana', 'apple',
                'sandwich', 'orange', 'broccoli', 'carrot', 'hot dog', 'pizza',
                'donut', 'cake', 'chair', 'couch', 'potted plant', 'bed',
                'dining table', 'toilet', 'tv', 'laptop', 'mouse', 'remote',
                'keyboard', 'cell phone', 'microwave', 'oven', 'toaster',
                'sink', 'refrigerator', 'book', 'clock', 'vase', 'scissors',
                'teddy bear', 'hair drier', 'toothbrush']

if __name__ == '__main__':
    """

    CommandLine:
        python -m geowatch.tasks.detectron2.predict
    """
    __cli__.main()



# coding=utf-8
# Copyright 2022 The Deeplab2 Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains functions to post-process MaX-DeepLab results."""

import functools
from typing import List, Tuple, Dict, Text

import tensorflow as tf
from deeplab2 import common
from deeplab2 import config_pb2
from deeplab2.data import dataset
from deeplab2.model import utils

_SMALL_CONSTANT_FOR_SOFTMAX = -99999


def _get_transformer_class_prediction_thing_stuff(
        transformer_class_probs: tf.Tensor,
        transformer_class_confidence_threshold_thing: float,
        transformer_class_confidence_threshold_stuff: float
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
    """Computes the transformer class prediction and confidence score.

    Args:
      transformer_class_probs: A tf.Tensor of shape [num_mask_slots,
        num_thing_stuff_classes + 1]. It is a pixel level logit scores where the
        num_mask_slots is the number of mask slots (for both thing classes and
        stuff classes) in MaX-DeepLab. The last channel indicates a `void` class.
      transformer_class_confidence_threshold_thing: A float for thresholding the
        confidence of the transformer_class_probs. The panoptic mask slots with
        class confidence less than the threshold are filtered and not used for
        panoptic prediction. Only masks whose confidence is larger than the
        threshold are counted in num_detections. This threshold applies to thing
        classes.
      transformer_class_confidence_threshold_stuff: A float for thresholding the
        confidence of the transformer_class_probs. The panoptic mask slots with
        class confidence less than the threshold are filtered and not used for
        panoptic prediction. Only masks whose confidence is larger than the
        threshold are counted in num_detections. This threshold applies to stuff
        classes.

    Returns:
      A tuple of:
      - the class prediction for all mask slots as float32 tf.Tensor of shape
        [num_mask_slots].
      - the class prediction confidence as float32 tf.Tensor of shape
        [num_mask_slots].
      - the number of detections as tf.Tensor of shape [1].
    """
    transformer_class_pred = tf.cast(
        tf.argmax(transformer_class_probs, axis=-1), tf.float32)
    transformer_class_confidence = tf.reduce_max(
        transformer_class_probs, axis=-1, keepdims=False)
    # Filter mask IDs with class confidence less than the threshold.
    thresholded_mask_thing = tf.cast(
        tf.greater_equal(transformer_class_confidence,
                         transformer_class_confidence_threshold_thing),
        tf.float32)
    thresholded_mask_stuff = tf.cast(
        tf.greater_equal(transformer_class_confidence,
                         transformer_class_confidence_threshold_stuff),
        tf.float32)
    num_detections = tf.reduce_sum(
        tf.cast(
            tf.greater(thresholded_mask_stuff + thresholded_mask_thing, 0.5),
            tf.int32))

    # Instead of gathering detected mask slots and return, we simply return all
    # class prediction with a binary mask for keep/drop to ensure static shape for
    # TPU running.
    return transformer_class_pred, transformer_class_confidence, num_detections


def _get_transformer_class_prediction(
        transformer_class_probs: tf.Tensor,
        transformer_class_confidence_threshold: float
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    """Computes the transformer class prediction and confidence score.

    Args:
      transformer_class_probs: A tf.Tensor of shape [num_mask_slots,
        num_thing_stuff_classes + 1]. It is a pixel level logit scores where the
        num_mask_slots is the number of mask slots (for both thing classes and
        stuff classes) in MaX-DeepLab. The last channel indicates a `void` class.
      transformer_class_confidence_threshold: A float for thresholding the
        confidence of the transformer_class_probs. The panoptic mask slots with
        class confidence less than the threshold are filtered and not used for
        panoptic prediction. Only masks whose confidence is larger than the
        threshold are counted in num_detections.

    Returns:
      A tuple of:
      - the class prediction for all mask slots as float32 tf.Tensor of shape
        [num_mask_slots].
      - the binary indicator of detected mask as tf.Tensor of shape
        [num_mask_slots].
      - the class prediction confidence as float32 tf.Tensor of shape
        [num_mask_slots].
      - the number of detections as tf.Tensor of shape [1].
    """
    transformer_class_pred = tf.cast(
        tf.argmax(transformer_class_probs, axis=-1), tf.float32)
    transformer_class_confidence = tf.reduce_max(
        transformer_class_probs, axis=-1, keepdims=False)
    # Filter mask IDs with class confidence less than the threshold.
    thresholded_mask = tf.cast(
        tf.greater_equal(transformer_class_confidence,
                         transformer_class_confidence_threshold), tf.float32)
    num_detections = tf.reduce_sum(
        tf.cast(tf.greater(thresholded_mask, 0.5), tf.int32))

    # Instead of gathering detected mask slots and return, we simply return all
    # class prediction with a binary mask for keep/drop to ensure static shape for
    # TPU running.
    return (transformer_class_pred, thresholded_mask,
            transformer_class_confidence, num_detections)


def _get_mask_id_and_semantic_maps(
        thing_class_ids: List[int],
        stuff_class_ids: List[int],
        pixel_space_mask_logits: tf.Tensor,
        transformer_class_probs: tf.Tensor,
        image_shape: List[int],
        pixel_confidence_threshold=0.4,
        transformer_class_confidence_threshold=0.7,
        transformer_post_processing='pixelwise',
        maskwise_postprocessing_config=None,
        pieces=1
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    """Computes the pixel-level mask ID map and semantic map per image or video.

     Some input and output tensors may contain num_frames in the shape, which is
     default to 1 (image-level processing).

    Args:
      thing_class_ids: A List of integers of shape [num_thing_classes] containing
        thing class indices.
      stuff_class_ids: A List of integers of shape [num_stuff_classes] containing
        stuff class indices.
      pixel_space_mask_logits: A tf.Tensor of shape [num_frames, height, width,
        num_mask_slots]. It is a pixel level logit scores where the num_mask_slots
        is the number of mask slots (for both thing classes and stuff classes) in
        MaX-DeepLab. num_frames is set to 1 by default (image-level processing).
      transformer_class_probs: A tf.Tensor of shape [num_mask_slots,
        num_thing_stuff_classes + 1]. It is a pixel level logit scores where the
        num_mask_slots is the number of mask slots (for both thing classes and
        stuff classes) in MaX-DeepLab. The last channel indicates a `void` class.
      image_shape: A list of integers specifying the [height, width] of input
        image.
      pixel_confidence_threshold: A float indicating a threshold for the pixel
        level softmax probability confidence of transformer mask logits. If less
        than the threshold, the pixel locations have confidence `0` in
        `confident_regions` output, and represent `void` (ignore) regions.
      transformer_class_confidence_threshold: A float for thresholding the
        confidence of the transformer_class_probs. The panoptic mask slots with
        class confidence less than the threshold are filtered and not used for
        panoptic prediction.
      transformer_post_processing: A string indicating which post-processing is
        used to obtain panoptic prediction. Currently two types of
        post-processings are supported: 'pixelwise' and 'maskwise'.
      maskwise_postprocessing_config: A dictionary of hyper-parameters for
        maskwise postprocessing.
      pieces: An integer indicating the number of pieces in the piece-wise
        operation. When computing panpotic prediction and confident regions, the
        mask logits are divided width-wise into multiple pieces and processed
        piece-wise due to the GPU memory limit. Then, the piece-wise outputs are
        concatenated along the width into the original mask shape. Defaults to 1.

    Returns:
      A tuple of:
      - the mask ID prediction as tf.Tensor of shape [num_frames, height, width].
      - the semantic prediction as tf.Tensor of shape [num_frames, height, width].
      - the thing region mask as tf.Tensor of shape [num_frames, height, width].
      - the stuff region mask as tf.Tensor of shape [num_frames, height, width].
      - the instance score map as tf.Tensor of shape [num_frames, height, width].
      - the semantic score map as tf.Tensor of shape [num_frames, height, width].

    Raises:
      ValueError: When input image's `width - 1` is not divisible by `pieces`.
      ValueError: When transformer_post_processing is not one of ['pixelwise',
        'maskwise'].
    """
    # The last channel indicates `void` class and thus is not included.
    transformer_class_probs = transformer_class_probs[..., :-1]
    # Generate mapping from mask IDs to dataset's thing and stuff semantic IDs.
    thing_stuff_class_ids = thing_class_ids + stuff_class_ids

    if transformer_post_processing == 'pixelwise':
        (transformer_class_pred, transformer_class_thresholded_mask,
         transformer_class_confidence, num_detections) = (
            _get_transformer_class_prediction(
                transformer_class_probs, transformer_class_confidence_threshold))
    elif transformer_post_processing == 'maskwise':
        transformer_class_pred, transformer_class_confidence, num_detections = (
            _get_transformer_class_prediction_thing_stuff(
                transformer_class_probs, maskwise_postprocessing_config[
                    'transformer_class_confidence_threshold_thing'],
                maskwise_postprocessing_config[
                    'transformer_class_confidence_threshold_stuff']))
    else:
        raise ValueError('transformer_post_processing must be one of [pixelwise,',
                         ' maskwise].')
    tf.assert_rank(
        pixel_space_mask_logits, 4,
        'pixel_space_mask_logits must have shape [num_frames,'
        ' height, width, num_mask_slots] and rank 4.')

    if transformer_post_processing == 'maskwise':
        # We obtain a mask confidence as average of probs of all assigned pixels.
        num_mask_slots = pixel_space_mask_logits.get_shape().as_list()[-1]
        pixel_space_mask_probs = tf.nn.softmax(pixel_space_mask_logits, axis=-1)
        detected_pixel_mask = tf.cast(
            pixel_space_mask_probs > pixel_confidence_threshold, tf.float32)
        mask_scores_flat = tf.reshape(pixel_space_mask_probs, [-1, num_mask_slots])
        detected_pixel_mask_flat = tf.reshape(detected_pixel_mask,
                                              [-1, num_mask_slots])
        mask_scores_flat = tf.reduce_sum(
            mask_scores_flat * detected_pixel_mask_flat, axis=0)
        mask_scores_flat = tf.math.divide_no_nan(
            mask_scores_flat, tf.reduce_sum(detected_pixel_mask_flat, axis=0))

        reorder_score = (
                                transformer_class_confidence **
                                maskwise_postprocessing_config['reorder_class_weight']
                        ) * (
                                mask_scores_flat **
                                maskwise_postprocessing_config['reorder_mask_weight'])

        mask_slots_reorder = tf.argsort(
            reorder_score, axis=-1, direction='DESCENDING', stable=True)
        # We re-order the predictions accordingly so they are arranged from more
        # confident to less.
        transformer_class_pred = tf.gather(
            transformer_class_pred, mask_slots_reorder, batch_dims=-1)
        transformer_class_confidence = tf.gather(
            transformer_class_confidence, mask_slots_reorder, batch_dims=-1)

    num_frames = pixel_space_mask_logits.get_shape().as_list()[0]

    # If num_detections = 0, return empty result maps.

    def _return_empty_mask_id_and_semantic_maps():
        return (tf.ones([num_frames, image_shape[0], image_shape[1]],
                        dtype=tf.int32),
                tf.zeros([num_frames, image_shape[0], image_shape[1]],
                         dtype=tf.int32),
                tf.zeros([num_frames, image_shape[0], image_shape[1]],
                         dtype=tf.float32),
                tf.zeros([num_frames, image_shape[0], image_shape[1]],
                         dtype=tf.float32),
                tf.zeros([num_frames, image_shape[0], image_shape[1]],
                         dtype=tf.float32),
                tf.zeros([num_frames, image_shape[0], image_shape[1]],
                         dtype=tf.float32))

    # If num_detections > 0 and transformer_post_processing == 'pixelwise':
    def _pixelwise_generate_mask_id_and_confident_maps():
        output_mask_id_map = []
        output_confident_region = []
        output_instance_score = []
        output_semantic_score = []
        logits_width = pixel_space_mask_logits.get_shape().as_list()[-2]
        output_width = image_shape[1]

        if (output_width - 1) % pieces > 0:
            raise ValueError('`output_width - 1` must be divisible by `pieces`.')
        # Use of input shape of a multiple of the feature stride, plus one, so that
        # it preserves left- and right-alignment.
        piece_output_width = (output_width - 1) // pieces + 1

        for piece_id in range(pieces):
            piece_begin = (logits_width - 1) // pieces * piece_id
            # Use of input shape of a multiple of the feature stride, plus one, so
            # that it preserves left- and right-alignment.
            piece_end = (logits_width - 1) // pieces * (piece_id + 1) + 1
            piece_pixel_mask_logits = (
                pixel_space_mask_logits[..., piece_begin:piece_end, :])
            piece_pixel_mask_logits = tf.compat.v1.image.resize_bilinear(
                piece_pixel_mask_logits, (image_shape[0], piece_output_width),
                align_corners=True)
            # Set logits of undetected mask slots to a very small number, which
            # will set the corresponding value after softmax to 0 thus has no impact.
            piece_detected_pixel_mask_logits = (
                    piece_pixel_mask_logits * transformer_class_thresholded_mask +
                    (_SMALL_CONSTANT_FOR_SOFTMAX *
                     (1 - transformer_class_thresholded_mask)))
            # Filter with pixel mask threshold.
            piece_pixel_confidence_map = tf.reduce_max(
                tf.nn.softmax(piece_detected_pixel_mask_logits, axis=-1), axis=-1)
            piece_confident_region = tf.cast(
                piece_pixel_confidence_map > pixel_confidence_threshold, tf.float32)
            # Filter the pixels which are assigned to a mask ID that does not survive.
            piece_max_logits = tf.reduce_max(piece_pixel_mask_logits, axis=-1)
            piece_detected_max_logits = tf.reduce_max(
                piece_detected_pixel_mask_logits, axis=-1)
            piece_detected_mask = tf.cast(
                tf.math.equal(piece_max_logits, piece_detected_max_logits),
                tf.float32)
            piece_confident_region = piece_confident_region * piece_detected_mask
            piece_mask_id_map = tf.cast(
                tf.argmax(piece_detected_pixel_mask_logits, axis=-1), tf.int32)
            piece_semantic_score = tf.gather(transformer_class_confidence,
                                             piece_mask_id_map)

            if piece_id == pieces - 1:
                output_mask_id_map.append(piece_mask_id_map)
                output_confident_region.append(piece_confident_region)
                output_instance_score.append(piece_pixel_confidence_map)
                output_semantic_score.append(piece_semantic_score)
            else:
                output_mask_id_map.append(piece_mask_id_map[..., :-1])
                output_confident_region.append(piece_confident_region[..., :-1])
                output_instance_score.append(piece_pixel_confidence_map[..., :-1])
                output_semantic_score.append(piece_semantic_score[..., :-1])
        return (output_mask_id_map, output_confident_region,
                output_instance_score, output_semantic_score)

    # If num_detections > 0 and transformer_post_processing == 'maskwise':
    def _maskwise_generate_mask_id_and_confident_maps():
        output_mask_id_map = []
        output_confident_region = []
        output_instance_score = []
        output_semantic_score = []
        logits_width = pixel_space_mask_logits.get_shape().as_list()[-2]
        output_width = image_shape[1]

        if (output_width - 1) % pieces > 0:
            raise ValueError('`output_width - 1` must be divisible by `pieces`.')
        # Use of input shape of a multiple of the feature stride, plus one, so that
        # it preserves left- and right-alignment.
        piece_output_width = (output_width - 1) // pieces + 1

        for piece_id in range(pieces):
            piece_begin = (logits_width - 1) // pieces * piece_id
            # Use of input shape of a multiple of the feature stride, plus one, so
            # that it preserves left- and right-alignment.
            piece_end = (logits_width - 1) // pieces * (piece_id + 1) + 1
            piece_pixel_mask_logits = (
                pixel_space_mask_logits[..., piece_begin:piece_end, :])
            piece_pixel_mask_logits = tf.compat.v1.image.resize_bilinear(
                piece_pixel_mask_logits, (image_shape[0], piece_output_width),
                align_corners=True)

            piece_pixel_mask_probs = tf.nn.softmax(piece_pixel_mask_logits, axis=-1)
            # In mask-wise post-processing, we firstly obtain a set of binary masks
            # where overlapping is allowed.
            piece_detected_pixel_mask = tf.cast(
                piece_pixel_mask_probs > pixel_confidence_threshold, tf.float32)
            # Similarly, the binary masks need to be re-ordered.
            piece_detected_pixel_mask = tf.gather(
                piece_detected_pixel_mask, mask_slots_reorder, batch_dims=-1)

            # We firstly generate an empty canvas, and paste the predicted masks onto
            # it in order.
            piece_mask_id_map = tf.zeros(
                [num_frames, image_shape[0], piece_output_width], dtype=tf.float32)
            piece_confident_region = tf.zeros(
                [num_frames, image_shape[0], piece_output_width], dtype=tf.float32)
            piece_instance_score = tf.reduce_max(piece_pixel_mask_probs, axis=-1)
            piece_semantic_score = tf.zeros(
                [num_frames, image_shape[0], piece_output_width], dtype=tf.float32)

            for mask_slots_idx in range(num_mask_slots):
                current_binary_mask = piece_detected_pixel_mask[..., mask_slots_idx]
                current_class_confidence = transformer_class_confidence[mask_slots_idx]
                current_class_pred = transformer_class_pred[mask_slots_idx]
                # We need to tell if current class prediction belongs to thing class or
                # stuff class, as different threshold may be used.
                is_current_class_pred_thing = tf.cast(
                    current_class_pred < len(thing_class_ids), tf.float32)
                is_current_class_pred_confident = (
                        tf.cast(
                            current_class_confidence >= maskwise_postprocessing_config[
                                'transformer_class_confidence_threshold_thing'], tf.float32)
                        * is_current_class_pred_thing + tf.cast(current_class_confidence >=
                                                                maskwise_postprocessing_config[
                                                                    'transformer_class_confidence_threshold_stuff'],
                                                                tf.float32) * (1 - is_current_class_pred_thing))
                # When copying current binary mask onto the canvas, some pixels may
                # already be occuipied by other masks with higher confidence, thus we
                # should remove these pixels from current binary mask. However, if too
                # many pixels are removed, we may also consider remove the whole mask.
                # maskwise_merge_overlap_threshold is used to decide when a whole mask
                # should be removed at this case.
                original_pixel_number = tf.reduce_sum(current_binary_mask)
                # We can only copy mask onto those unoccupied pixels.
                new_binary_mask = current_binary_mask * (
                    tf.cast(piece_mask_id_map == 0, tf.float32))
                new_pixel_number = tf.reduce_sum(new_binary_mask)
                current_mask_not_overlap_too_much = tf.cast(
                    new_pixel_number > original_pixel_number *
                    maskwise_postprocessing_config['overlapping_threshold'], tf.float32)
                # We keep new_binary_mask if it is confident and do not overlap too
                # much with previous masks.
                new_binary_mask = (
                        new_binary_mask * is_current_class_pred_confident *
                        current_mask_not_overlap_too_much)
                # Paste the mask onto canvas.
                piece_mask_id_map = (
                        piece_mask_id_map +
                        new_binary_mask * tf.cast(mask_slots_idx + 1, tf.float32))
                piece_confident_region = (piece_confident_region + new_binary_mask)
                piece_semantic_score = ((1.0 - new_binary_mask) * piece_semantic_score
                                        + new_binary_mask * current_class_confidence)

            # At final, we minus one so the mask_id will correctly start from 0. Note
            # that the unconfident pixels may be wrongly classified to mask slot 0,
            # but it is fine as it will finally be filtered by confident_region_mask.
            # In pixelwise post-processing, the piece_mask_id_map will range from 0
            # to N-1, and unconfident pixels will still have the mask_id based on
            # argmax, which is filtered based on confident_region variable afterwards.
            # Here as we use maskwise post-processing, we do not know unconfident
            # pixels' "correct" assginment based on argmax, thus we simply assgin
            # them all to mask slot 0, which is filtered based on confident_region
            # variable afterwards. This can also ensure that piece_mask_id_map will
            # range from 0 to N-1.
            piece_mask_id_map = piece_mask_id_map - 1
            piece_mask_id_map = tf.maximum(piece_mask_id_map, 0)
            piece_mask_id_map = tf.cast(piece_mask_id_map, tf.int32)

            if piece_id == pieces - 1:
                output_mask_id_map.append(piece_mask_id_map)
                output_confident_region.append(piece_confident_region)
                output_instance_score.append(piece_instance_score)
                output_semantic_score.append(piece_semantic_score)
            else:
                output_mask_id_map.append(piece_mask_id_map[..., :-1])
                output_confident_region.append(piece_confident_region[..., :-1])
                output_instance_score.append(piece_instance_score[..., :-1])
                output_semantic_score.append(piece_semantic_score[..., :-1])
        return (output_mask_id_map, output_confident_region,
                output_instance_score, output_semantic_score)

    def _generate_mask_id_and_semantic_maps():
        (output_mask_id_map, output_confident_region, output_instance_score,
         output_semantic_score) = (
            _pixelwise_generate_mask_id_and_confident_maps()
            if transformer_post_processing == 'pixelwise' else
            _maskwise_generate_mask_id_and_confident_maps())
        mask_id_map = tf.concat(output_mask_id_map, axis=-1)
        confident_region = tf.concat(output_confident_region, axis=-1)
        instance_score_map = tf.concat(output_instance_score, axis=-1)
        semantic_score_map = tf.concat(output_semantic_score, axis=-1)
        mask_id_map_flat = tf.reshape(mask_id_map, [-1])
        mask_id_semantic_map_flat = tf.gather(transformer_class_pred,
                                              mask_id_map_flat)
        mask_id_semantic_map = tf.reshape(
            mask_id_semantic_map_flat, [num_frames, image_shape[0], image_shape[1]])

        # Generate thing and stuff masks (with value 1/0 indicates the
        # presence/absence)
        thing_mask = tf.cast(mask_id_semantic_map < len(thing_class_ids),
                             tf.float32) * confident_region
        stuff_mask = tf.cast(mask_id_semantic_map >= len(thing_class_ids),
                             tf.float32) * confident_region
        # Generate semantic_map.
        semantic_map = tf.gather(
            tf.convert_to_tensor(thing_stuff_class_ids),
            tf.cast(tf.round(mask_id_semantic_map_flat), tf.int32))
        semantic_map = tf.reshape(semantic_map,
                                  [num_frames, image_shape[0], image_shape[1]])
        # Add 1 because mask ID 0 is reserved for unconfident region.
        mask_id_map_plus_one = mask_id_map + 1
        semantic_map = tf.cast(tf.round(semantic_map), tf.int32)
        return (mask_id_map_plus_one, semantic_map, thing_mask, stuff_mask,
                instance_score_map, semantic_score_map)

    (mask_id_map_plus_one, semantic_map, thing_mask, stuff_mask,
     instance_score_map, semantic_score_map) = tf.cond(
        tf.cast(num_detections, tf.float32) < tf.cast(0.5, tf.float32),
        _return_empty_mask_id_and_semantic_maps,
        _generate_mask_id_and_semantic_maps)

    return (mask_id_map_plus_one, semantic_map, thing_mask, stuff_mask,
            instance_score_map, semantic_score_map)


def _filter_by_count(input_index_map: tf.Tensor,
                     area_limit: int) -> Tuple[tf.Tensor, tf.Tensor]:
    """Filters input index map by area limit threshold per index.

    Args:
      input_index_map: A float32 tf.Tensor of shape [batch, num_frames, height,
        width].
      area_limit: An integer specifying the number of pixels that each index
        regions need to have at least. If not over the limit, the index regions
        are masked (zeroed) out.

    Returns:
      masked input_index_map: A tf.Tensor with shape [batch, num_frames, height,
        width], masked by the area_limit threshold.
      mask: A tf.Tensor with shape [batch, num_frames, height, width]. It is a
        pixel-level mask with 1. indicating the regions over the area limit, and
        0. otherwise.
    """
    input_h, input_w = input_index_map.get_shape().as_list()[-2:]
    index_map = tf.cast(tf.round(input_index_map), tf.int32)
    index_map_flat = tf.reshape(index_map, [-1, input_h * input_w])
    counts = tf.math.bincount(index_map_flat, axis=-1)
    counts_map = tf.gather(counts, index_map_flat, batch_dims=1)
    counts_map = tf.reshape(counts_map, tf.shape(index_map))

    mask = tf.cast(
        tf.cast(counts_map, tf.float32) > tf.cast(area_limit - 0.5, tf.float32),
        input_index_map.dtype)
    return input_index_map * mask, mask


def _merge_mask_id_and_semantic_maps(
        mask_id_maps_plus_one: tf.Tensor,
        semantic_maps: tf.Tensor,
        thing_masks: tf.Tensor,
        stuff_masks: tf.Tensor,
        void_label: int,
        label_divisor: int,
        thing_area_limit: int,
        stuff_area_limit: int,
) -> tf.Tensor:
    """Merges mask_id maps and semantic_maps to obtain panoptic segmentation.

    Args:
      mask_id_maps_plus_one: A tf.Tensor of shape [batch, height, width] or
        [batch, num_frames, height, width].
      semantic_maps: A tf.Tensor of shape [batch, height, width] or [batch,
        num_frames, height, width].
      thing_masks: A float32 tf.Tensor of shape [batch, height, width] or [batch,
        num_frames, height, width] containing masks with 1. at thing regions, 0.
        otherwise.
      stuff_masks: A float32 tf.Tensor of shape [batch, height, width] or [batch,
        num_frames, height, width] containing masks with 1. at thing regions, 0.
        otherwise.
      void_label: An integer specifying the void label.
      label_divisor: An integer specifying the label divisor of the dataset.
      thing_area_limit: An integer specifying the number of pixels that thing
        regions need to have at least. The thing region will be included in the
        panoptic prediction, only if its area is larger than the limit; otherwise,
        it will be re-assigned as void_label. A function to filter out thing
        regions smaller than thing_area_limit is applied when thing_area_limit is
        larger than 0.
      stuff_area_limit: An integer specifying the number of pixels that stuff
        regions need to have at least. The stuff region will be included in the
        panoptic prediction, only if its area is larger than the limit; otherwise,
        it will be re-assigned as void_label. A function to filter out stuff
        regions smaller than stuff_area_limit is applied when stuff_area_limit is
        larger than 0.

    Returns:
      panoptic_maps: A tf.Tensor with shape [batch, num_frames, height, width].
    """
    thing_mask_id_maps_plus_one = (
            tf.cast(mask_id_maps_plus_one, tf.float32) * thing_masks)
    # We increase semantic_maps by 1 before masking (zeroing) by thing_masks and
    # stuff_masks, to ensure all valid semantic IDs are greater than 0 and thus
    # not masked out.
    semantic_maps_plus_one = semantic_maps + 1
    tf.debugging.assert_less(
        tf.reduce_sum(thing_masks * stuff_masks),
        0.5,
        message='thing_masks and stuff_masks must be mutually exclusive.')

    thing_semantic_maps = (
            tf.cast(semantic_maps_plus_one, tf.float32) * thing_masks)
    stuff_semantic_maps = (
            tf.cast(semantic_maps_plus_one, tf.float32) * stuff_masks)
    if stuff_area_limit > 0:
        # Filter stuff_semantic_maps by stuff_area_limit.
        stuff_semantic_maps, _ = _filter_by_count(stuff_semantic_maps,
                                                  stuff_area_limit)
    if thing_area_limit > 0:
        # Filter thing_mask_id_map and thing_semantic_map by thing_area_limit
        thing_mask_id_maps_plus_one, mask_id_count_filter_mask = _filter_by_count(
            thing_mask_id_maps_plus_one, thing_area_limit)
        thing_semantic_maps = thing_semantic_maps * mask_id_count_filter_mask

    # Filtered un-confident region will be replaced with `void_label`. The
    # "plus_one" will be reverted, the un-confident region (0) will be -1, and so
    # we add (void + 1)
    semantic_maps_new = thing_semantic_maps + stuff_semantic_maps - 1.0
    semantic_maps_new = (
            tf.cast(semantic_maps_new < -0.5, tf.float32) *
            tf.cast(void_label + 1, tf.float32) + semantic_maps_new)
    panoptic_maps = (
            semantic_maps_new * label_divisor + thing_mask_id_maps_plus_one)
    panoptic_maps = tf.cast(tf.round(panoptic_maps), tf.int32)
    return panoptic_maps


@tf.function
def _get_panoptic_predictions(
        pixel_space_mask_logits: tf.Tensor,
        transformer_class_logits: tf.Tensor,
        thing_class_ids: List[int],
        void_label: int,
        label_divisor: int,
        thing_area_limit: int,
        stuff_area_limit: int,
        image_shape: List[int],
        pixel_confidence_threshold=0.4,
        transformer_class_confidence_threshold=0.7,
        transformer_post_processing='pixelwise',
        maskwise_postprocessing_config=None,
        pieces=1
) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]:
    """Computes the pixel-level panoptic, mask ID, and semantic maps.

    Args:
      pixel_space_mask_logits: A tf.Tensor of shape [batch, num_frames,
        strided_height, strided_width, num_mask_slots]. It is a pixel level logit
        scores where the num_mask_slots is the number of mask slots (for both
        thing classes and stuff classes) in MaX-DeepLab.
      transformer_class_logits: A tf.Tensor of shape [batch, num_mask_slots,
        num_thing_stuff_classes + 1]. It is a pixel level logit scores where the
        num_mask_slots is the number of mask slots (for both thing classes and
        stuff classes) in MaX-DeepLab. The last channel indicates a `void` class.
      thing_class_ids: A List of integers of shape [num_thing_classes] containing
        thing class indices.
      void_label: An integer specifying the void label.
      label_divisor: An integer specifying the label divisor of the dataset.
      thing_area_limit: An integer specifying the number of pixels that thing
        regions need to have at least. The thing region will be included in the
        panoptic prediction, only if its area is larger than the limit; otherwise,
        it will be re-assigned as void_label.
      stuff_area_limit: An integer specifying the number of pixels that stuff
        regions need to have at least. The stuff region will be included in the
        panoptic prediction, only if its area is larger than the limit; otherwise,
        it will be re-assigned as void_label.
      image_shape: A list of integers specifying the [height, width] of input
        image.
      pixel_confidence_threshold: A float indicating a threshold for the pixel
        level softmax probability confidence of transformer mask logits. If less
        than the threshold, the pixel locations have confidence `0` in
        `confident_regions` output, and represent `void` (ignore) regions.
      transformer_class_confidence_threshold: A float for thresholding the
        confidence of the transformer_class_probs. The panoptic mask slots with
        class confidence less than the threshold are filtered and not used for
        panoptic prediction.
      transformer_post_processing: A string indicating which post-processing is
        used to obtain panoptic prediction. Currently two types of
        post-processings are supported: 'pixelwise' and 'maskwise'.
      maskwise_postprocessing_config: A dictionary of hyper-parameters for
        maskwise postprocessing.
      pieces: An integer indicating the number of pieces in the piece-wise
        operation in `_get_mask_id_and_semantic_maps`. When computing panoptic
        prediction and confident regions, the mask logits are divided width-wise
        into multiple pieces and processed piece-wise due to the GPU memory limit.
        Then, the piece-wise outputs are concatenated along the width into the
        original mask shape. Defaults to 1.

    Returns:
      A tuple of:
      - the panoptic prediction as tf.Tensor with shape [batch, num_frames,
          height, width].
      - the mask ID prediction as tf.Tensor with shape [batch, num_frames, height,
          width].
      - the semantic prediction as tf.Tensor with shape [batch, num_frames,
          height, width].
      - the instance score map as tf.Tensor with shape [batch, num_frames,
          height, width].
      - the semantic score map as tf.Tensor with shape [batch, num_frames,
          height, width].
    """
    transformer_class_probs = tf.nn.softmax(transformer_class_logits, axis=-1)
    batch_size = utils.resolve_batch_size(transformer_class_logits)
    # num_thing_stuff_classes does not include `void` class, so we decrease by 1.
    num_thing_stuff_classes = (
            transformer_class_logits.get_shape().as_list()[-1] - 1)
    # Generate thing and stuff class ids
    stuff_class_ids = utils.get_stuff_class_ids(num_thing_stuff_classes,
                                                thing_class_ids, void_label)

    mask_id_map_plus_one_lists = tf.TensorArray(
        tf.int32, size=batch_size, dynamic_size=False)
    semantic_map_lists = tf.TensorArray(
        tf.int32, size=batch_size, dynamic_size=False)
    thing_mask_lists = tf.TensorArray(
        tf.float32, size=batch_size, dynamic_size=False)
    stuff_mask_lists = tf.TensorArray(
        tf.float32, size=batch_size, dynamic_size=False)
    instance_score_map_lists = tf.TensorArray(
        tf.float32, size=batch_size, dynamic_size=False)
    semantic_score_map_lists = tf.TensorArray(
        tf.float32, size=batch_size, dynamic_size=False)
    for i in tf.range(batch_size):
        (mask_id_map_plus_one, semantic_map, thing_mask, stuff_mask,
         instance_score_map, semantic_score_map) = (
            _get_mask_id_and_semantic_maps(thing_class_ids, stuff_class_ids,
                                           pixel_space_mask_logits[i, ...],
                                           transformer_class_probs[i, ...],
                                           image_shape, pixel_confidence_threshold,
                                           transformer_class_confidence_threshold,
                                           transformer_post_processing,
                                           maskwise_postprocessing_config, pieces))
        mask_id_map_plus_one_lists = mask_id_map_plus_one_lists.write(
            i, mask_id_map_plus_one)
        semantic_map_lists = semantic_map_lists.write(i, semantic_map)
        thing_mask_lists = thing_mask_lists.write(i, thing_mask)
        stuff_mask_lists = stuff_mask_lists.write(i, stuff_mask)
        instance_score_map_lists = (
            instance_score_map_lists.write(i, instance_score_map))
        semantic_score_map_lists = (
            semantic_score_map_lists.write(i, semantic_score_map))
    # This does not work with unknown shapes.
    mask_id_maps_plus_one = mask_id_map_plus_one_lists.stack()
    semantic_maps = semantic_map_lists.stack()
    thing_masks = thing_mask_lists.stack()
    stuff_masks = stuff_mask_lists.stack()
    instance_score_maps = instance_score_map_lists.stack()
    semantic_score_maps = semantic_score_map_lists.stack()

    panoptic_maps = _merge_mask_id_and_semantic_maps(
        mask_id_maps_plus_one, semantic_maps, thing_masks, stuff_masks,
        void_label, label_divisor, thing_area_limit, stuff_area_limit)
    return (panoptic_maps, mask_id_maps_plus_one, semantic_maps,
            instance_score_maps, semantic_score_maps)


class PostProcessor(tf.keras.layers.Layer):
    """This class contains code of a MaX-DeepLab post-processor."""

    def __init__(self,
                 config: config_pb2.ExperimentOptions,
                 dataset_descriptor: dataset.DatasetDescriptor,
                 name: str = 'PostProcessor'):
        """Initializes a MaX-DeepLab post-processor.

        We currently support two different ways for post-processing:
          - pixel-wise [1]: In pixel-wise post-processing, the panoptic prediction
            is obtained with two argmax, where the first argmax decicdes the
            assignment from pixels to mask slots, and second one decides the
            predicted semantic class of each mask slot.

          - mask-wise [2,3,4]: In mask-wise post-processing, the panoptic prediction
            is obtained in a copy-paste manner. Specifically, a set of binary masks
            is obtained for each mask slot by thresholding. Afterwards, the binary
            masks are re-ordered with a confidence score based on class
            confidence and/or mask confidence. Then the masks are copied to an
            empty canvas in the order from high confidence to low confidence.

        References:
          [1] MaX-DeepLab: End-to-End Panoptic Segmentation with Mask Transformers,
              CVPR 2021.
                Huiyu Wang, Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.

          [2] Per-Pixel Classification is Not All You Need for Semantic
              Segmentation, NeurIPS 2021.
                Bowen Cheng, Alexander G. Schwing, Alexander Kirillov.

          [3] Panoptic SegFormer: Delving Deeper into Panoptic Segmentation with
              Transformers, CVPR 2022.
                Zhiqi Li, Wenhai Wang, Enze Xie, Zhiding Yu, Anima Anandkumar,
                Jose M. Alvarez, Ping Luo, Tong Lu.

          [4] CMT-DeepLab: Clustering Mask Transformers for Panoptic Segmentation,
              CVPR 2022.
                Qihang Yu, Huiyu Wang, Dahun Kim, Siyuan Qiao, Maxwell Collins,
                Yukun Zhu, Hartwig Adam, Alan Yuille, Liang-Chieh Chen.

        Args:
          config: A config_pb2.ExperimentOptions configuration.
          dataset_descriptor: A dataset.DatasetDescriptor.
          name: Name of the layer.
        """
        super(PostProcessor, self).__init__(name=name)
        # Convert maskwise_postprocessing_config to a dict.
        maskwise_postprocessing_config = None
        if config.evaluator_options.HasField('maskwise_postprocessing'):
            maskwise_postprocessing_config = {
                'transformer_class_confidence_threshold_thing':
                    config.evaluator_options.maskwise_postprocessing
                    .transformer_class_confidence_threshold_thing,
                'transformer_class_confidence_threshold_stuff':
                    config.evaluator_options.maskwise_postprocessing
                    .transformer_class_confidence_threshold_stuff,
                'overlapping_threshold':
                    config.evaluator_options.maskwise_postprocessing
                    .overlapping_threshold,
                'reorder_class_weight':
                    config.evaluator_options.maskwise_postprocessing
                    .reorder_class_weight,
                'reorder_mask_weight':
                    config.evaluator_options.maskwise_postprocessing
                    .reorder_mask_weight,
            }

        self._post_processor = functools.partial(
            _get_panoptic_predictions,
            thing_class_ids=list(dataset_descriptor.class_has_instances_list),
            void_label=dataset_descriptor.ignore_label,
            label_divisor=dataset_descriptor.panoptic_label_divisor,
            thing_area_limit=config.evaluator_options.thing_area_limit,
            stuff_area_limit=config.evaluator_options.stuff_area_limit,
            image_shape=list(config.eval_dataset_options.crop_size),
            pixel_confidence_threshold=config.evaluator_options
            .pixel_confidence_threshold,
            transformer_class_confidence_threshold=config.evaluator_options
            .transformer_class_confidence_threshold,
            transformer_post_processing=config.evaluator_options
            .transformer_post_processing,
            maskwise_postprocessing_config=maskwise_postprocessing_config,
            pieces=1)

        # We use another post-processor for semantic prediction and evaluation,
        # where we do not apply any thresholding. When run panoptic model for
        # semantic segmentation, we set the threshold to 0 for a better results
        # (as semantic label does not prefer 'void'). With another post-processor
        # where no thresholding is applied, we do not need to run the model twice,
        # and we may have more accurate evaluation of mIoU on the fly.
        self._post_processor_semantic = functools.partial(
            _get_panoptic_predictions,
            thing_class_ids=list(dataset_descriptor.class_has_instances_list),
            void_label=dataset_descriptor.ignore_label,
            label_divisor=dataset_descriptor.panoptic_label_divisor,
            thing_area_limit=0,
            stuff_area_limit=0,
            image_shape=list(config.eval_dataset_options.crop_size),
            pixel_confidence_threshold=0,
            transformer_class_confidence_threshold=0,
            transformer_post_processing='pixelwise',
            maskwise_postprocessing_config=None,
            pieces=1)

        # Post-processing for video data is not fully supported yet, so we fix the
        # num_frames to 1.
        self._num_frames = 1

    def _expand_num_frames_dimension(self, result_dict):
        """Expands the num_frames dimension of pixel predictions."""
        pixel_space = result_dict[common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY]
        pixel_shape = pixel_space.get_shape().as_list()
        # Expand the num_frames dimension as [batch * num_frames, height, width,
        # num_mask_slots] -> [batch, num_frames, height, width, num_mask_slots].
        if len(pixel_shape) == 4:
            result_dict[common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY] = tf.reshape(
                pixel_space, [-1, self._num_frames] + pixel_shape[1:])
        return result_dict

    def _merge_batch_and_num_frames_dimensions(self, result_dict):
        """Merges the batch and num_frames dimensions of the pixel predictions."""
        for key in result_dict.keys():
            # Squeeze [batch, num_frames, height, width] -> [batch * num_frames,
            # height, width]
            result_shape = result_dict[key].get_shape().as_list()
            if len(result_shape) >= 4:
                result_dict[key] = tf.reshape(result_dict[key], [-1] + result_shape[2:])
        return result_dict

    def call(self, result_dict: Dict[Text, tf.Tensor]) -> Dict[Text, tf.Tensor]:
        """Performs the post-processing given model predicted results.

        Args:
          result_dict: A dictionary of tf.Tensor containing model results. The dict
            has to contain:
              - common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY,
              - common.PRED_TRANSFORMER_CLASS_LOGITS_KEY,

        Returns:
          The post-processed dict of tf.Tensor, containing the following:
           - common.PRED_SEMANTIC_KEY,
           - common.PRED_INSTANCE_KEY,
           - common.PRED_PANOPTIC_KEY,
        """
        # Make a copy so that input argument will not be modified, per requirements
        # from exporting a saved model.
        result_dict = dict(result_dict)

        # We handle both image [batch, height, width, ...] and video [batch,
        # num_frames, height, width, ...] inputs. To unify the input and output
        # shape for all the post-processing functions, we expand the num_frames
        # dimension for image-level tensors (num_frames = 1) before calling the
        # post-process function. Then, we merge the batch and num_frames dimensions.
        result_dict = self._expand_num_frames_dimension(result_dict)
        processed_dict = {}
        (processed_dict[common.PRED_PANOPTIC_KEY],
         processed_dict[common.PRED_INSTANCE_KEY],
         _,
         processed_dict[common.PRED_INSTANCE_SCORES_KEY],
         processed_dict[common.PRED_SEMANTIC_SCORES_KEY]) = self._post_processor(
            result_dict[common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY],
            result_dict[common.PRED_TRANSFORMER_CLASS_LOGITS_KEY])

        # We obtain semantic predictions with another post-processor, where
        # thresholding is not applied.
        (_,
         _,
         processed_dict[common.PRED_SEMANTIC_KEY],
         _,
         _) = self._post_processor_semantic(
            result_dict[common.PRED_PIXEL_SPACE_MASK_LOGITS_KEY],
            result_dict[common.PRED_TRANSFORMER_CLASS_LOGITS_KEY])

        processed_dict = self._merge_batch_and_num_frames_dimensions(processed_dict)

        return processed_dict



def demo():
    """
    Show how to get the predicted heatmaps in video space at a resolution for
    each frame in a video. Also show how to get the corresponding rasterized
    annotations.
    """
    import watch
    import kwimage
    import numpy as np

    if 1:
        coco_dset = watch.coerce_kwcoco('watch-msi', heatmap=True, geodata=True)

    chosen_video = coco_dset.videos().objs[0]
    video_id = chosen_video['id']
    images_for_video = coco_dset.images(video_id=video_id)

    resolution = '10GSD'

    def get_annots_in_requested_space(coco_img, space='video', resolution=None):
        """
        This is slightly less than ideal in terms of API, but it will work for
        now.
        """
        # Build transform from image to requested space
        warp_vid_from_img = coco_img._warp_vid_from_img
        scale = coco_img._scalefactor_for_resolution(space='video',
                                                     resolution=resolution,
                                                     RESOLUTION_KEY='target_gsd')
        warp_req_from_vid = kwimage.Affine.scale(scale)
        warp_req_from_img = warp_req_from_vid @ warp_vid_from_img

        # Get annotations in "Image Space"
        annots = coco_img.dset.annots(image_id=coco_img.img['id'])
        imgspace_dets = annots.detections
        reqspace_dets = imgspace_dets.warp(warp_req_from_img)
        return reqspace_dets

    frame_data_list = []
    frame_truth_list = []

    for coco_img in images_for_video.coco_images:
        # Get the image data in "requested space"
        delayed = coco_img.imdelay('salient', space='video', resolution=resolution, RESOLUTION_KEY='target_gsd')
        detections = get_annots_in_requested_space(coco_img, space='video', resolution=resolution)

        frame_data = delayed.finalize()

        h, w = frame_data.shape[0:2]

        # TODO: filter by class or whatever you want here
        truth_canvas = np.zeros((h, w))
        detections.data['segmentations'].fill(truth_canvas, value=1)
        frame_truth_list.append(truth_canvas)
        frame_data_list.append(frame_data)



from kwcoco_dataloader.tasks.fusion.datamodules.network_io import BatchItem  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import HeterogeneousBatchItem  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import HomogeneousBatchItem  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import RGBImageBatchItem  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import _demo_dataset  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import UncollatedBatch  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import HeterogeneousBatch  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import UncollatedRGBImageBatch  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import CollatedBatch  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import CollatedRGBImageBatch  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import NetworkOutputs  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import UncollatedNetworkOutputs  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import CollatedNetworkOutputs  # NOQA
from kwcoco_dataloader.tasks.fusion.datamodules.network_io import decollate  # NOQA



from torch.nn import Conv2d, Linear
from torch import where, rand, topk, long, empty, zeros, no_grad, tensor
import torch
import sys
from lop.utils.AdamGnT import AdamGnT
from torch.nn.init import calculate_gain
from lop.utils.miscellaneous import get_layer_bound


class ConvGnT(object):
    """
    Generate-and-Test algorithm for ConvNets, maturity threshold based tester, accumulates probability of replacement,
    with various measures of feature utility
    """
    def __init__(self, net, hidden_activation, opt, decay_rate=0.99, replacement_rate=1e-4, init='kaiming',
                 num_last_filter_outputs=4, util_type='contribution', maturity_threshold=100, device='cpu'):
        super(ConvGnT, self).__init__()

        self.net = net
        self.num_hidden_layers = int(len(self.net)/2)
        self.util_type = util_type
        self.device = device

        self.opt = opt
        self.opt_type = 'sgd'
        if isinstance(self.opt, AdamGnT):
            self.opt_type = 'AdamGnT'

        """
        Define the hyper-parameters of the algorithm
        """
        self.replacement_rate = replacement_rate
        self.decay_rate = decay_rate
        self.num_last_filter_outputs = num_last_filter_outputs
        self.maturity_threshold = maturity_threshold
        self.util_type = util_type

        """
        Utility of all features/neurons
        """
        self.util, self.bias_corrected_util, self.ages, self.mean_feature_act, self.mean_abs_feature_act, \
             = [], [], [], [], []

        for i in range(self.num_hidden_layers):
            if isinstance(self.net[i * 2], Conv2d):
                self.util.append(zeros(self.net[i * 2].out_channels))
                self.bias_corrected_util.append(zeros(self.net[i * 2].out_channels))
                self.ages.append(zeros(self.net[i * 2].out_channels))
                self.mean_feature_act.append(zeros(self.net[i * 2].out_channels))
                self.mean_abs_feature_act.append(zeros(self.net[i * 2].out_channels))
            elif isinstance(self.net[i * 2], Linear):
                self.util.append(zeros(self.net[i * 2].out_features))
                self.bias_corrected_util.append(zeros(self.net[i * 2].out_features))
                self.ages.append(zeros(self.net[i * 2].out_features))
                self.mean_feature_act.append(zeros(self.net[i * 2].out_features))
                self.mean_abs_feature_act.append(zeros(self.net[i * 2].out_features))

        self.accumulated_num_features_to_replace = [0 for i in range(self.num_hidden_layers)]
        self.m = torch.nn.Softmax(dim=1)

        """
        Calculate uniform distribution's bound for random feature initialization
        """
        if hidden_activation == 'selu': init = 'lecun'
        self.bounds = self.compute_bounds(hidden_activation=hidden_activation, init=init)
        """
        Pre calculate number of features to replace per layer per update
        """
        self.num_new_features_to_replace = []
        for i in range(self.num_hidden_layers):
            with no_grad():
                if isinstance(self.net[i * 2], Linear):
                    self.num_new_features_to_replace.append(self.replacement_rate * self.net[i * 2].out_features)
                elif isinstance(self.net[i * 2], Conv2d):
                    self.num_new_features_to_replace.append(self.replacement_rate * self.net[i * 2].out_channels)

    def compute_bounds(self, hidden_activation, init='kaiming'):
        if hidden_activation in ['swish', 'elu']: hidden_activation = 'relu'
        bounds = []
        gain = calculate_gain(nonlinearity=hidden_activation)
        for i in range(self.num_hidden_layers):
            bounds.append(get_layer_bound(layer=self.net[i * 2], init=init, gain=gain))
        bounds.append(get_layer_bound(layer=self.net[-1], init=init, gain=1))
        return bounds

    def update_utility(self, layer_idx=0, features=None):
        with torch.no_grad():
            self.util[layer_idx] *= self.decay_rate
            bias_correction = 1 - self.decay_rate ** self.ages[layer_idx]

            current_layer = self.net[layer_idx * 2]
            next_layer = self.net[layer_idx * 2 + 2]

            if isinstance(next_layer, Linear):
                output_wight_mag = next_layer.weight.data.abs().mean(dim=0)
            elif isinstance(next_layer, Conv2d):
                output_wight_mag = next_layer.weight.data.abs().mean(dim=(0, 2, 3))

            self.mean_feature_act[layer_idx] *= self.decay_rate
            self.mean_abs_feature_act[layer_idx] *= self.decay_rate
            if isinstance(current_layer, Linear):
                input_wight_mag = current_layer.weight.data.abs().mean(dim=1)
                self.mean_feature_act[layer_idx] += (1 - self.decay_rate) * features.mean(dim=0)
                self.mean_abs_feature_act[layer_idx] += (1 - self.decay_rate) * features.abs().mean(dim=0)
            elif isinstance(current_layer, Conv2d):
                input_wight_mag = current_layer.weight.data.abs().mean(dim=(1, 2, 3))
                if isinstance(next_layer, Conv2d):
                    self.mean_feature_act[layer_idx] += (1 - self.decay_rate) * features.mean(dim=(0, 2, 3))
                    self.mean_abs_feature_act[layer_idx] += (1 - self.decay_rate) * features.abs().mean(dim=(0, 2, 3))
                else:
                    self.mean_feature_act[layer_idx] += (1 - self.decay_rate) * features.mean(dim=0).view(-1, self.num_last_filter_outputs).mean(dim=1)
                    self.mean_abs_feature_act[layer_idx] += (1 - self.decay_rate) * features.abs().mean(dim=0).view(-1, self.num_last_filter_outputs).mean(dim=1)

            bias_corrected_act = self.mean_feature_act[layer_idx] / bias_correction

            if self.util_type == 'adaptation':
                new_util = 1 / input_wight_mag
            elif self.util_type in ['contribution', 'zero_contribution', 'adaptable_contribution']:
                if self.util_type == 'contribution':
                    bias_corrected_act = 0
                else:
                    if isinstance(current_layer, Conv2d):
                        if isinstance(next_layer, Conv2d):
                            bias_corrected_act = bias_corrected_act.view(1, -1, 1, 1)
                        else:
                            bias_corrected_act = bias_corrected_act.repeat_interleave(self.num_last_filter_outputs).view(1, -1)
                if isinstance(next_layer, Linear):
                    if isinstance(current_layer, Linear):
                        new_util = output_wight_mag * (features - bias_corrected_act).abs().mean(dim=0)
                    elif isinstance(current_layer, Conv2d):
                        new_util = (output_wight_mag * (features - bias_corrected_act).abs().mean(dim=0)).view(-1, self.num_last_filter_outputs).mean(dim=1)
                elif isinstance(next_layer, Conv2d):
                    new_util = output_wight_mag * (features - bias_corrected_act).abs().mean(dim=(0, 2, 3))
                if self.util_type == 'adaptable_contribution':
                    new_util = new_util / input_wight_mag

            if self.util_type == 'random':
                self.bias_corrected_util[layer_idx] = rand(self.util[layer_idx].shape)
            else:
                self.util[layer_idx] += (1 - self.decay_rate) * new_util
                # correct the bias in the utility computation
                self.bias_corrected_util[layer_idx] = self.util[layer_idx] / bias_correction

    def test_features(self, features):
        """
        Args:
            features: Activation values in the neural network
        Returns:
            Features to replace in each layer, Number of features to replace in each layer
        """
        features_to_replace_input_indices = [empty(0, dtype=long) for _ in range(self.num_hidden_layers)]
        features_to_replace_output_indices = [empty(0, dtype=long) for _ in range(self.num_hidden_layers)]
        num_features_to_replace = [0 for _ in range(self.num_hidden_layers)]
        if self.replacement_rate == 0:
            return features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace

        for i in range(self.num_hidden_layers):
            self.ages[i] += 1
            """
            Update feature utility
            """
            self.update_utility(layer_idx=i, features=features[i])
            """
            Find the no. of features to replace
            """
            eligible_feature_indices = where(self.ages[i] > self.maturity_threshold)[0]
            if eligible_feature_indices.shape[0] == 0:
                continue
            self.accumulated_num_features_to_replace[i] += self.num_new_features_to_replace[i]

            """
            Case when the number of features to be replaced is between 0 and 1.
            """
            num_new_features_to_replace = int(self.accumulated_num_features_to_replace[i])
            self.accumulated_num_features_to_replace[i] -= num_new_features_to_replace

            if num_new_features_to_replace == 0:    continue

            """
            Find features to replace in the current layer
            """
            new_features_to_replace = topk(-self.bias_corrected_util[i][eligible_feature_indices],
                                           num_new_features_to_replace)[1]
            new_features_to_replace = eligible_feature_indices[new_features_to_replace]

            """
            Initialize utility for new features
            """
            self.util[i][new_features_to_replace] = 0
            self.mean_feature_act[i][new_features_to_replace] = 0.
            self.mean_abs_feature_act[i][new_features_to_replace] = 0.

            num_features_to_replace[i] = num_new_features_to_replace
            features_to_replace_input_indices[i] = new_features_to_replace
            features_to_replace_output_indices[i] = new_features_to_replace
            if isinstance(self.net[i * 2], Conv2d) and isinstance(self.net[i * 2 + 2], Linear):
                features_to_replace_output_indices[i] = \
                    (new_features_to_replace*self.num_last_filter_outputs).repeat_interleave(self.num_last_filter_outputs) + \
                    tensor([i for i in range(self.num_last_filter_outputs)]).repeat(new_features_to_replace.size()[0])

        return features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace

    def update_optim_params(self, features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace):
        """
        Update Optimizer's state
        """
        if self.opt_type == 'AdamGnT':
            for i in range(self.num_hidden_layers):
                # input weights
                if num_features_to_replace == 0:
                    continue
                # input weights
                self.opt.state[self.net[i * 2].bias]['exp_avg'][features_to_replace_input_indices[i]] = 0.0
                self.opt.state[self.net[i * 2].weight]['exp_avg_sq'][features_to_replace_input_indices[i], :] = 0.0
                self.opt.state[self.net[i * 2].bias]['exp_avg_sq'][features_to_replace_input_indices[i]] = 0.0
                self.opt.state[self.net[i * 2].weight]['step'][features_to_replace_input_indices[i], :] = 0
                self.opt.state[self.net[i * 2].bias]['step'][features_to_replace_input_indices[i]] = 0
                # output weights
                self.opt.state[self.net[i * 2 + 2].weight]['exp_avg'][:, features_to_replace_output_indices[i]] = 0.0
                self.opt.state[self.net[i * 2 + 2].weight]['exp_avg_sq'][:, features_to_replace_output_indices[i]] = 0.0
                self.opt.state[self.net[i * 2 + 2].weight]['step'][:, features_to_replace_output_indices[i]] = 0

    def gen_new_features(self, features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace):
        """
        Generate new features: Reset input and output weights for low utility features
        """
        with torch.no_grad():
            for i in range(self.num_hidden_layers):
                if num_features_to_replace[i] == 0:
                    continue
                current_layer = self.net[i * 2]
                next_layer = self.net[i * 2 + 2]

                if isinstance(current_layer, Linear):
                    current_layer.weight.data[features_to_replace_input_indices[i], :] *= 0.0
                    current_layer.weight.data[features_to_replace_input_indices[i], :] -= - \
                        empty(num_features_to_replace[i], current_layer.in_features).uniform_(-self.bounds[i],
                                                                                                self.bounds[i]).to(self.device)
                elif isinstance(current_layer, Conv2d):
                    current_layer.weight.data[features_to_replace_input_indices[i], :] *= 0.0
                    current_layer.weight.data[features_to_replace_input_indices[i], :] -= - \
                        empty([num_features_to_replace[i]] + list(current_layer.weight.shape[1:])). \
                            uniform_(-self.bounds[i], self.bounds[i])

                current_layer.bias.data[features_to_replace_input_indices[i]] *= 0.0
                """
                # Set the outgoing weights and ages to zero
                """
                next_layer.weight.data[:, features_to_replace_output_indices[i]] = 0
                self.ages[i][features_to_replace_input_indices[i]] = 0

    def gen_and_test(self, features):
        """
        Perform generate-and-test
        :param features: activation of hidden units in the neural network
        """
        if not isinstance(features, list):
            print('features passed to generate-and-test should be a list')
            sys.exit()
        features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace = self.test_features(features=features)
        self.gen_new_features(features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace)
        self.update_optim_params(features_to_replace_input_indices, features_to_replace_output_indices, num_features_to_replace)
